from pathlib import Path
import json
from ._arcgis_model import _EmptyData, _change_tail, ArcGISModel, _get_device
from ._codetemplate import code, image_classifier_prf, panoptic_segmenter_prf
import warnings
import arcgis
import sys, os, importlib
from functools import partial
import logging

logger = logging.getLogger()

HAS_OPENCV = True
HAS_FASTAI = True

try:
    import torch
    from torch import nn
    import numpy as np
    from fastai.basic_train import Learner, LearnerCallback
    from fastai.torch_core import split_model_idx
    from fastai.vision import ImageList
    from fastai.vision import imagenet_stats, normalize
    from fastai.core import has_arg, split_kwargs_by_func
    from fastai.basic_data import DatasetType
    from fastai.callback import Callback
    from fastai.torch_core import to_cpu, grab_idx
    from fastai.basic_train import loss_batch
    from fastai.vision.image import open_image, bb2hw, image2np, Image, pil2tensor
    from .._utils.classified_tiles import per_class_metrics
    from ._deeplab_utils import compute_miou
    import PIL
    from ._ssd_utils import compute_class_AP
    from .._utils.pascal_voc_rectangles import (
        show_results_multispectral,
        ObjectDetectionCategoryList,
    )
    from ._unet_utils import (
        show_results_multispectral as show_results_multispectral_segmentation,
    )
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from ._arcgis_model import _set_ddp_multigpu, _isnotebook
    from ._hed_utils import accuracies
    from .._image_utils import (
        _get_image_chips,
        _get_transformed_predictions,
        _draw_predictions,
        _exclude_detection,
    )
    from .._video_utils import VideoUtils
    import inspect
    from .._utils.env import is_arcgispronotebook
    from matplotlib import pyplot as plt
    import types
    from ._maskrcnn import grid_anchors
    from .._utils.pascal_voc_rectangles import _reconstruct
    from .._utils.utils import chips_to_batch

    HAS_FASTAI = True

except Exception as e:
    HAS_FASTAI = False

try:
    import cv2
except Exception:
    HAS_OPENCV = False


class ModelExtension(ArcGISModel):
    """
    Creates a ModelExtension object, to train the model for object detection, semantic segmentation, and edge detection.

    =====================   ============================================================
    **Parameter**            **Description**
    ---------------------   ------------------------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            :meth:`~arcgis.learn.prepare_data`  function.
    ---------------------   ------------------------------------------------------------
    model_conf              A class definition contains the following methods:

                            * ``get_model(self, data, backbone=None, **kwargs)``: for model definition,

                            * ``on_batch_begin(self, learn, model_input_batch, model_target_batch, **kwargs)``: for feeding input to the model during training,

                            * ``transform_input(self, xb)``: for feeding input to the model during inferencing/validation,

                            * ``transform_input_multispectral(self, xb)``: for feeding input to the model during inferencing/validation in case of multispectral data,

                            * ``loss(self, model_output, *model_target)``: to return loss value of the model

                            * ``post_process(self, pred, nms_overlap, thres, chip_size, device)``: to post-process
                              the output of the object-detection model.

                            * ``post_process(self, pred, thres)``: to post-process the output of the segmentation model.
    ---------------------   ------------------------------------------------------------
    backbone                Optional function. If custom model requires any backbone.
    ---------------------   ------------------------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    =====================   ============================================================

    :return: :class:`~arcgis.learn.ModelExtension` Object

    """

    def __init__(self, data, model_conf, backbone=None, pretrained_path=None, **kwargs):
        self._learn_version = kwargs.get("ArcGISLearnVersion", "1.9.1")

        if self._learn_version >= "1.9.0":
            if pretrained_path is not None:
                pretrained_backbone = False
            else:
                pretrained_backbone = True
            kwargs["pretrained_backbone"] = pretrained_backbone
        else:
            del kwargs["ArcGISLearnVersion"]

        super().__init__(data, backbone, pretrained_path=pretrained_path, **kwargs)
        data = self._data
        self._model_conf = model_conf()
        self._model_conf_class = model_conf
        self._backend = "pytorch"
        self._kwargs = kwargs
        model = self._model_conf.get_model(data, backbone, **kwargs)
        if backbone is not None:
            backbone_name = backbone if type(backbone) is str else backbone.__name__
            if model_conf.__name__ == "MyFasterRCNN" and backbone_name not in [
                "resnet50",
                "resnet101",
                "resnet152",
            ]:
                model.rpn.anchor_generator.grid_anchors = types.MethodType(
                    grid_anchors, model.rpn.anchor_generator
                )
        if self._is_multispectral:
            model = _change_tail(model, data)
        if not _isnotebook():
            _set_ddp_multigpu(self)
            if self._multigpu_training:
                self.learn = Learner(
                    data, model, loss_func=self._model_conf.loss
                ).to_distributed(self._rank_distributed)
                self._map_location = {"cuda:%d" % 0: "cuda:%d" % self._rank_distributed}
            else:
                self.learn = Learner(data, model, loss_func=self._model_conf.loss)
        else:
            self.learn = Learner(data, model, loss_func=self._model_conf.loss)
        self.learn.callbacks.append(
            self._train_callback(self.learn, self._model_conf.on_batch_begin)
        )
        self.learn._learn_version = self._learn_version
        if self._data.dataset_type == "Classified_Tiles":
            if getattr(self, "_is_edge_detection", False):
                from ._hed_utils import accuracy, f1_score

                self.learn.metrics = [accuracy, f1_score]
            else:
                from ._psp_utils import accuracy

                ignore_class = kwargs.get("ignore_class", [])
                accuracy = partial(accuracy, ignore_mapped_class=ignore_class)
                self.learn.metrics = [accuracy]
            self._code = image_classifier_prf
        elif self._data.dataset_type == "Panoptic_Segmentation":
            self._code = panoptic_segmenter_prf
            self._kwargs["n_masks"] = self._data.K
            self._kwargs["instance_classes"] = self._data.instance_classes
        else:
            self._code = code
        self._arcgis_init_callback()  # make first conv weights learnable
        self._bind_dataset_methods()
        if pretrained_path is not None:
            self.load(pretrained_path)

    if HAS_FASTAI:

        class _train_callback(LearnerCallback):
            def __init__(self, learn, on_batch_begin_fn):
                super().__init__(learn)
                self.on_batch_begin_fn = on_batch_begin_fn

            def on_batch_begin(self, last_input, last_target, **kwargs):
                if self._learn_version >= "1.9.0":
                    last_input, last_target = self.on_batch_begin_fn(
                        self.learn, last_input, last_target, **kwargs
                    )
                else:
                    last_input, last_target = self.on_batch_begin_fn(
                        self.learn, last_input, last_target
                    )

                return {"last_input": last_input, "last_target": last_target}

    def _analyze_pred(
        self, pred, thresh=0.5, nms_overlap=0.1, ret_scores=True, device=None
    ):
        return self._model_conf.post_process(
            pred, nms_overlap, thresh, self.learn.data.chip_size, device
        )

    def _get_emd_params(self, save_inference_file):
        import random

        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        if self._data.dataset_type == "Classified_Tiles":
            _emd_template["ModelType"] = "ImageClassification"
            if save_inference_file:
                _emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
            else:
                _emd_template[
                    "InferenceFunction"
                ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageClassifier.py"
            _emd_template["IsEdgeDetection"] = getattr(
                self, "_is_edge_detection", False
            )
            _emd_template["ModelConfiguration"] = "_model_extension_inferencing"

        elif self._data.dataset_type == "Panoptic_Segmentation":
            _emd_template["ModelType"] = "PanopticSegmenter"
            if save_inference_file:
                _emd_template["InferenceFunction"] = "ArcGISPanopticSegmenter.py"
            else:
                _emd_template[
                    "InferenceFunction"
                ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISPanopticSegmenter.py"
            _emd_template["ModelConfiguration"] = "_panoptic_inferencing"

        else:
            _emd_template["ModelType"] = "ObjectDetection"
            if save_inference_file:
                _emd_template["InferenceFunction"] = "ArcGISObjectDetector.py"
            else:
                _emd_template[
                    "InferenceFunction"
                ] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISObjectDetector.py"
            _emd_template["ModelConfiguration"] = "_model_extension_inferencing"
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template["Classes"] = []
        _emd_template["ModelConfigurationFile"] = "ModelConfiguration.py"
        _emd_template["ModelFileConfigurationClass"] = type(self._model_conf).__name__
        _emd_template["DatasetType"] = self._data.dataset_type
        if not hasattr(self._data.train_dl.dataset, "coco"):
            _emd_template["Kwargs"] = self._kwargs

        class_data = {}
        for i, class_name in enumerate(
            self._data.classes[1:]
        ):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            _emd_template["Classes"].append(class_data.copy())

        return _emd_template

    @property
    def _is_model_extension(self):
        return True

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a :class:`~arcgis.learn.ModelExtension` object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from :meth:`~arcgis.learn.prepare_data`  function or None for
                                inferencing.
        =====================   ===========================================

        :return: :class:`~arcgis.learn.ModelExtension` Object
        """

        emd_path = _get_emd_path(emd_path)

        with open(emd_path) as f:
            emd = json.load(f)

        model_file = Path(emd["ModelFile"])

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        modelconf = Path(emd["ModelConfigurationFile"])

        if not modelconf.is_absolute():
            modelconf = emd_path.parent / modelconf

        modelconfclass = emd["ModelFileConfigurationClass"]

        spec = importlib.util.spec_from_file_location(modelconfclass, modelconf)
        module = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = module
        spec.loader.exec_module(module)
        model_configuration = getattr(module, modelconfclass)

        backbone = emd["ModelParameters"].get("backbone", None)

        dataset_type = emd.get("DatasetType", "PASCAL_VOC_rectangles")
        chip_size = emd["ImageWidth"]
        resize_to = emd.get("resize_to", None)
        kwargs = emd.get("Kwargs", {})
        kwargs["ArcGISLearnVersion"] = emd.get("ArcGISLearnVersion", "1.0.0")
        if isinstance(resize_to, list):
            resize_to = (resize_to[0], resize_to[1])

        try:
            class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}
            color_mapping = {i["Value"]: i["Color"] for i in emd["Classes"]}
        except KeyError:
            class_mapping = {i["ClassValue"]: i["ClassName"] for i in emd["Classes"]}
            color_mapping = {i["ClassValue"]: i["Color"] for i in emd["Classes"]}

        data_passed = True
        if data is None:
            data_passed = False
            if dataset_type == "PASCAL_VOC_rectangles":
                train_tfms = []
                val_tfms = []
                ds_tfms = (train_tfms, val_tfms)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)
                    sd = ImageList([], path=emd_path.parent.parent).split_by_idx([])
                    data = (
                        sd.label_const(
                            0,
                            label_cls=ObjectDetectionCategoryList,
                            classes=list(class_mapping.values()),
                        )
                        .transform(ds_tfms)
                        .databunch(device=_get_device())
                        .normalize(imagenet_stats)
                    )
                # Add 1 for background class
                data.c += 1
            else:
                data = _EmptyData(
                    path=emd_path.parent.parent,
                    loss_func=None,
                    c=len(class_mapping) + 1,
                    chip_size=emd["ImageHeight"],
                )

            data.chip_size = chip_size
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data.classes = ["background"] + list(class_mapping.values())
            data._is_empty = True
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)
            data.dataset_type = dataset_type
            if dataset_type == "Panoptic_Segmentation":
                data.K = emd["Kwargs"]["n_masks"]
                data.instance_classes = emd["Kwargs"]["instance_classes"]
        data.resize_to = resize_to

        mextnsn = cls(
            data,
            model_configuration,
            backbone,
            pretrained_path=str(model_file),
            **kwargs,
        )

        if not data_passed and dataset_type == "PASCAL_VOC_rectangles":
            mextnsn.learn.data.single_ds.classes = mextnsn._data.classes
            mextnsn.learn.data.single_ds.y.classes = mextnsn._data.classes

        return mextnsn

    @property
    def _model_metrics(self):
        if self._data.dataset_type == "Classified_Tiles":
            return {"accuracy": "{0:1.4e}".format(self._get_model_metrics())}
        elif self._data.dataset_type == "Panoptic_Segmentation":
            pq = self.panoptic_quality(show_progress=False)
            return {"panoptic_quality": "{:.4f}".format(pq)}
        else:
            return {
                "average_precision_score": self.average_precision_score(
                    show_progress=False
                )
            }

    def _get_model_metrics(self, **kwargs):
        checkpoint = getattr(self, "_is_checkpointed", False)
        if not hasattr(self.learn, "recorder"):
            return 0.0

        if len(self.learn.recorder.metrics) == 0:
            return 0.0

        model_accuracy = self.learn.recorder.metrics[-1][0]
        if checkpoint:
            model_accuracy = self.learn.recorder.metrics[self.learn._best_epoch][0]
        return float(model_accuracy)

    def _get_y(self, bbox, clas):
        try:
            bbox = bbox.view(-1, 4)
        except Exception:
            bbox = torch.zeros(size=[0, 4])
        bb_keep = ((bbox[:, 2] - bbox[:, 0]) > 0).nonzero()[:, 0]
        return bbox[bb_keep], clas[bb_keep]

    def _intersect(self, box_a, box_b):
        max_xy = torch.min(box_a[:, None, 2:], box_b[None, :, 2:])
        min_xy = torch.max(box_a[:, None, :2], box_b[None, :, :2])
        inter = torch.clamp((max_xy - min_xy), min=0)
        return inter[:, :, 0] * inter[:, :, 1]

    def _box_sz(self, b):
        return (b[:, 2] - b[:, 0]) * (b[:, 3] - b[:, 1])

    def _jaccard(self, box_a, box_b):
        inter = self._intersect(box_a, box_b)
        union = (
            self._box_sz(box_a).unsqueeze(1) + self._box_sz(box_b).unsqueeze(0) - inter
        )
        return inter / union

    def _bind_dataset_methods(self):
        if self._data.dataset_type == "Classified_Tiles":
            if getattr(self, "_is_edge_detection", False):
                self.show_results = self._show_results_edge_detection
                self.compute_precision_recall = self._edge_detection_accuracies
            else:
                if self._is_multispectral:
                    self.show_results = self._show_results_multispectral_segmentation
                else:
                    self.show_results = self._show_results_segmentation
                self.mIOU = self._mIOU
                self.per_class_metrics = self._per_class_metrics
                self.accuracy = self._accuracy

        elif self._data.dataset_type == "Panoptic_Segmentation":
            self.show_results = self._show_results_panoptic
            self.panoptic_quality = self._panoptic_quality

        else:
            if self._is_multispectral:
                self.show_results = self._show_results_multispectral
            else:
                self.show_results = self._show_results_object_detection
            self.average_precision_score = self._average_precision_score
            self.predict = self._predict
            self.predict_video = self._predict_video

    def _accuracy(self):
        """
        Returns accuracy of the model.
        """
        try:
            return self.learn.validate()[1].tolist()
        except Exception as e:
            accuracy = self._data.emd.get("accuracy")
            if accuracy:
                return accuracy
            else:
                logger.error("Metric not found in the loaded model")

    def _mIOU(self, mean=False, show_progress=True):
        """
        Computes mean IOU on the validation set for each class.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                mean IOU, otherwise returns mean iou of all
                                classes combined.
        ---------------------   -------------------------------------------
        show_progress           Optional bool. Displays the progress bar if
                                True.
        =====================   ===========================================
        :return: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()
        num_classes = torch.arange(self._data.c)
        miou = compute_miou(self, self._data.valid_dl, mean, num_classes, show_progress)
        if mean:
            return np.mean(miou)
        return dict(zip(["0"] + self._data.classes[1:], miou))

    def _per_class_metrics(self, ignore_classes=[]):
        """
        Computer per class precision, recall and f1-score on validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        self                    segmentation model object -> [PSPNetClassifier | UnetClassifier | DeepLab]
        ---------------------   -------------------------------------------
        ignore_classes          Optional list. It will contain the list of class
                            values on which model will not incur loss.
                            Default: []
        =====================   ===========================================

        Returns per class precision, recall and f1 scores
        """
        # Standalone models will be missing _ignore_classes attribute
        if hasattr(self, "_ignore_classes"):
            ignore_classes = np.unique(self._ignore_classes + ignore_classes).tolist()
        else:
            ignore_classes = np.unique(ignore_classes).tolist()

        try:
            self._check_requisites()
            ## Calling imported function `per_class_metrics`
            return per_class_metrics(self, ignore_classes)
        except:
            import pandas as pd

            return pd.read_json(self._data.emd["per_class_metrics"])

    def _show_results_object_detection(self, rows=5, thresh=0.5, nms_overlap=0.1):
        """
        Displays the results of a trained model on a part of the validation set.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        =====================   ===========================================
        """
        self._check_requisites()
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)
        self._show_results_modified(
            rows=rows, thresh=thresh, nms_overlap=nms_overlap, model=self
        )

    def _show_results_segmentation(self, rows=5, thresh=0.5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional Float. The probability above which
                                a detection will be considered valid.
        =====================   ===========================================
        """
        self._check_requisites()
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)

        self._show_results_modified(rows=rows, thresh=thresh, model=self, **kwargs)

    def _show_results_panoptic(self, rows=5, thresh=0.5, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional Float. The probability above which
                                a detection will be considered valid.
        =====================   ===========================================
        **kwargs**
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        alpha                   Optional Float. Default value is 0.5.
                                Opacity of the lables for the corresponding
                                images. Values range between 0 and 1, where
                                1 means opaque.
        ---------------------   -------------------------------------------
        random_colors           Optional Boolean. Default value is False.
                                Assigns randomized colors to each class and
                                instance.
        =====================   ===========================================
        """
        self._check_requisites()
        # Limit the number of rows to validation dataset length
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)

        from ._max_deeplab_utils import show_results_panoptic

        show_results_panoptic(self, rows=rows, thresh=thresh, **kwargs)

    def _show_results_edge_detection(self, rows=5, thresh=0.5, thinning=True, **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.
        """
        self._check_requisites()
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)

        self._show_results_modified(
            rows=rows, thresh=thresh, model=self, thinning=thinning, **kwargs
        )

    def _show_results_multispectral(
        self, rows=5, thresh=0.3, nms_overlap=0.1, alpha=1, **kwargs
    ):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional int. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        alpha                   Optional Float.
                                Opacity of the lables for the corresponding
                                images. Values range between 0 and 1, where
                                1 means opaque.
        =====================   ===========================================

        """
        ret_val = show_results_multispectral(
            self,
            nrows=rows,
            thresh=thresh,
            nms_overlap=nms_overlap,
            alpha=alpha,
            **kwargs,
        )
        if kwargs.get("return_fig", False):
            fig, ax = ret_val
            return fig

    def _show_results_multispectral_segmentation(
        self, rows=5, alpha=0.7, **kwargs
    ):  # parameters adjusted in kwargs
        """
        Displays the results of a trained model on a part of the validation set.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer. Number of rows of results
                                to be displayed.
        ---------------------   -------------------------------------------
        alpha                   Optional Float.
                                Opacity of the lables for the corresponding
                                images. Values range between 0 and 1, where
                                1 means opaque.
        =====================   ===========================================
        """
        return_fig = kwargs.get("return_fig", False)
        ret_val = show_results_multispectral_segmentation(
            self, nrows=rows, alpha=alpha, **kwargs
        )
        if return_fig:
            fig, ax = ret_val
            return fig

    def _show_results_modified(self, rows=5, **kwargs):
        if rows > len(self._data.valid_ds):
            rows = len(self._data.valid_ds)

        ds_type = DatasetType.Valid
        n_items = rows**2 if self.learn.data.train_ds.x._square_show_res else rows
        if self.learn.dl(ds_type).batch_size < n_items:
            n_items = self.learn.dl(ds_type).batch_size
        ds = self.learn.dl(ds_type).dataset
        xb, yb = self.learn.data.one_batch(ds_type, detach=False, denorm=False)
        self.learn.model.eval()
        transform_kwargs, kwargs = split_kwargs_by_func(
            kwargs, self._model_conf.transform_input
        )
        try:
            preds = self.learn.model(
                self._model_conf.transform_input(xb, **transform_kwargs)
            )
        except Exception as e:
            if getattr(self, "_is_fasterrcnn", False):
                preds = []
                for _ in range(xb.shape[0]):
                    res = {}
                    res["boxes"] = torch.empty(0, 4)
                    res["scores"] = torch.tensor([])
                    res["labels"] = torch.tensor([])
                    preds.append(res)
            else:
                raise e

        x, y = to_cpu(xb), to_cpu(yb)
        norm = getattr(self.learn.data, "norm", False)
        if norm:
            x = self.learn.data.denorm(x)
            if norm.keywords.get("do_y", False):
                y = self.learn.data.denorm(y, do_x=True)
                preds = self.learn.data.denorm(preds, do_x=True)
        analyze_kwargs, kwargs = split_kwargs_by_func(kwargs, ds.y.analyze_pred)
        preds = ds.y.analyze_pred(preds, **analyze_kwargs)
        xs = [ds.x.reconstruct(grab_idx(x, i)) for i in range(n_items)]
        if has_arg(ds.y.reconstruct, "x"):
            ys = [ds.y.reconstruct(grab_idx(y, i), x=x) for i, x in enumerate(xs)]
            zs = [ds.y.reconstruct(z, x=x) for z, x in zip(preds, xs)]
        else:
            ys = [ds.y.reconstruct(grab_idx(y, i)) for i in range(n_items)]
            zs = [ds.y.reconstruct(z) for z in preds]
        ds.x.show_xyzs(xs, ys, zs, **kwargs)
        if is_arcgispronotebook():
            plt.show()

    def _predict_learn_modified(self, item, **kwargs):
        "Return predicted class, label and probabilities for `item`."
        batch = item
        transform_kwargs, kwargs = split_kwargs_by_func(
            kwargs, self._model_conf.transform_input
        )
        self.learn.model.eval()
        try:
            pred = self.learn.model(
                self._model_conf.transform_input(batch, **transform_kwargs)
            )
        except Exception as e:
            if getattr(self, "_is_fasterrcnn", False):
                pred = []
                for _ in range(batch[0].shape[0]):
                    res = {}
                    res["boxes"] = torch.empty(0, 4)
                    res["scores"] = torch.tensor([])
                    res["labels"] = torch.tensor([])
                    pred.append(res)
            else:
                raise e
        ds = self.learn.data.single_ds
        analyze_kwargs, kwargs = split_kwargs_by_func(kwargs, ds.y.analyze_pred)
        pred = ds.y.analyze_pred(pred, **analyze_kwargs)
        return pred

    def _average_precision_score(
        self, detect_thresh=0.2, iou_thresh=0.1, mean=False, show_progress=True
    ):
        """
        Computes average precision on the validation set for each class.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth labels, above
                                which a predicted bounding box will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        =====================   ===========================================
        :return: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()

        aps = compute_class_AP(
            self,
            self._data.valid_dl,
            self._data.c - 1,
            show_progress,
            iou_thresh=iou_thresh,
            detect_thresh=detect_thresh,
            thresh=detect_thresh,
            nms_overlap=iou_thresh,
        )
        if mean:
            import statistics

            return statistics.mean(aps)
        else:
            return dict(zip(self._data.classes[1:], aps))

    def _edge_detection_accuracies(self, thresh=0.5, buffer=3, show_progress=True):
        """
        Computes precision, recall and f1 score on validation set.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        thresh                  Optional float. The probability above which
                                a detection will be considered edge pixel.
        ---------------------   -------------------------------------------
        buffer                  Optional int. pixels in neighborhood to
                                consider true detection.
        =====================   ===========================================
        :return: `dict`
        """
        self._check_requisites()
        acc = accuracies(
            self,
            self._data.valid_dl,
            detect_thresh=thresh,
            buffer=buffer,
            show_progress=show_progress,
        )
        return acc

    def _panoptic_quality(self, show_progress=True, **kwargs):
        """
        Computes the Panoptic Quality metric for panoptic segmentation.
         =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        show_progress           Optional bool. Displays the progress bar if
                                True.
        =====================   ===========================================
        :return: `float`
        """
        from ._max_deeplab_utils import compute_panoptic_quality

        pq = compute_panoptic_quality(self, show_progress, **kwargs)
        return pq

    def _get_batched_predictions(
        self, chips, tytx, threshold, nms_overlap, normalize, batch_size=1
    ):
        data = []
        data_counter = 0
        final_output = []

        for idx, chip in enumerate(chips):
            frame = np.moveaxis(
                normalize(
                    cv2.cvtColor(chip["chip"], cv2.COLOR_BGR2RGB).astype(np.float32)
                ),
                -1,
                0,
            )
            data.append(frame)
            data_counter += 1
            if data_counter % batch_size == 0 or idx == len(chips) - 1:
                batch = chips_to_batch(data, tytx, tytx, batch_size)
                extra_chips = batch_size - len(data)
                batch_final = batch[: (len(batch) - extra_chips)]
                predictions = self._predict_learn_modified(
                    torch.tensor(batch_final).float().to(self._device),
                    thresh=threshold,
                    nms_overlap=nms_overlap,
                    ret_scores=True,
                    model=self,
                )
                ds = self.learn.data.single_ds
                for e, prediction in enumerate(predictions):
                    prediction = [
                        prediction[0].detach().cpu(),
                        prediction[1].detach().cpu(),
                        prediction[2].detach().cpu(),
                    ]
                    batch_img = torch.tensor(batch_final[e])
                    norm = getattr(self.learn.data, "norm", False)
                    if norm:
                        batch_img = self.learn.data.denorm(batch_img)
                    x = ds.x.reconstruct(grab_idx(batch_img, 0))
                    y = (
                        ds.y.reconstruct(prediction, x)
                        if has_arg(ds.y.reconstruct, "x")
                        else ds.y.reconstruct(prediction)
                    )
                    final_output.append(y)
                data = []
                data_counter = 0

        return final_output

    def _predict(
        self,
        image_path,
        threshold=0.5,
        nms_overlap=0.1,
        return_scores=False,
        visualize=False,
        resize=False,
        batch_size=1,
    ):
        """
        Runs prediction on an Image.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        image_path              Required. Path to the image file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        return_scores           Optional boolean. Will return the probability
                                scores of the bounding box predictions if True.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. Displays the image with
                                predicted bounding boxes if True.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the image to the same size
                                (chip_size parameter in prepare_data) that the model
                                was trained on, before detecting objects.
                                Note that if resize_to parameter was used in prepare_data,
                                the image is resized to that size instead.
                                By default, this parameter is false and the detections are
                                run in a sliding window fashion by applying the model on
                                cropped sections of the image (of the same size as the
                                model was trained on).
        ---------------------   -------------------------------------------
        batch_size              Optional int. Batch size to be used
                                during tiled inferencing. Deafult value 1.
        ---------------------   -------------------------------------------
        =====================   ===========================================

        :return:  Returns a tuple with predictions, labels and optionally confidence scores
                   if return_scores=True. The predicted bounding boxes are returned as a list
                   of lists containing the  xmin, ymin, width and height of each predicted
                   object in each image. The labels are returned as a list of class values
                   and the confidence scores are returned as a list of floats indicating
                   the confidence of each prediction.
        """
        if not HAS_OPENCV:
            raise Exception(
                "This function requires opencv 4.0.1.24. Install it using pip install opencv-python==4.0.1.24"
            )

        if isinstance(image_path, str):
            image = cv2.imread(image_path)
        else:
            image = image_path

        if image is None:
            raise Exception(str("No such file or directory: %s" % (image_path)))

        orig_height, orig_width, _ = image.shape
        orig_frame = image.copy()

        if resize and self._data.resize_to is None and self._data.chip_size is not None:
            image = cv2.resize(image, (self._data.chip_size, self._data.chip_size))

        if self._data.resize_to is not None:
            if isinstance(self._data.resize_to, tuple):
                image = cv2.resize(image, self._data.resize_to)
            else:
                image = cv2.resize(image, (self._data.resize_to, self._data.resize_to))

        height, width, _ = image.shape
        tytx = self._data.chip_size

        if self._data.chip_size is not None:
            chips = _get_image_chips(image, self._data.chip_size)
        else:
            chips = [
                {
                    "width": width,
                    "height": height,
                    "xmin": 0,
                    "ymin": 0,
                    "chip": image,
                    "predictions": [],
                }
            ]

        include_pad_detections = False
        if len(chips) == 1:
            include_pad_detections = True

        imagenet_stats = ([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])

        mean = 255 * np.array(imagenet_stats[0], dtype=np.float32)
        std = 255 * np.array(imagenet_stats[1], dtype=np.float32)
        norm = lambda x: (x - mean) / std

        valid_tfms = self._data.valid_ds.tfms
        self._data.valid_ds.tfms = []

        from .._utils.pascal_voc_rectangles import modified_getitem
        from fastai.data_block import LabelList

        orig_getitem = LabelList.__getitem__
        LabelList.__getitem__ = modified_getitem
        try:
            prediction_data = self._get_batched_predictions(
                chips, tytx, threshold, nms_overlap, norm, batch_size
            )
            for idx, bbox in enumerate(prediction_data):
                if bbox:
                    scores = bbox.scores
                    bboxes, lbls = bbox._compute_boxes()
                    bboxes.add_(1).mul_(
                        torch.tensor(
                            [
                                chips[idx]["height"] / 2,
                                chips[idx]["width"] / 2,
                                chips[idx]["height"] / 2,
                                chips[idx]["width"] / 2,
                            ]
                        )
                    ).long()
                    for index, bbox in enumerate(bboxes):
                        if lbls is not None:
                            label = lbls[index]
                        else:
                            label = "Default"

                        data = bb2hw(bbox)
                        if include_pad_detections or not _exclude_detection(
                            (data[0], data[1], data[2], data[3]),
                            chips[idx]["width"],
                            chips[idx]["height"],
                        ):
                            chips[idx]["predictions"].append(
                                {
                                    "xmin": data[0],
                                    "ymin": data[1],
                                    "width": data[2],
                                    "height": data[3],
                                    "score": float(scores[index]),
                                    "label": label,
                                }
                            )
        finally:
            LabelList.__getitem__ = orig_getitem

        self._data.valid_ds.tfms = valid_tfms

        predictions, labels, scores = _get_transformed_predictions(chips)

        y_ratio = orig_height / height
        x_ratio = orig_width / width

        for index, prediction in enumerate(predictions):
            prediction[0] = prediction[0] * x_ratio
            prediction[1] = prediction[1] * y_ratio
            prediction[2] = prediction[2] * x_ratio
            prediction[3] = prediction[3] * y_ratio

            # Clip xmin
            if prediction[0] < 0:
                prediction[2] = prediction[2] + prediction[0]
                prediction[0] = 1

            # Clip width when xmax greater than original width
            if prediction[0] + prediction[2] > orig_width:
                prediction[2] = (prediction[0] + prediction[2]) - orig_width

            # Clip ymin
            if prediction[1] < 0:
                prediction[3] = prediction[3] + prediction[1]
                prediction[1] = 1

            # Clip height when ymax greater than original height
            if prediction[1] + prediction[3] > orig_height:
                prediction[3] = (prediction[1] + prediction[3]) - orig_height

            predictions[index] = [
                prediction[0],
                prediction[1],
                prediction[2],
                prediction[3],
            ]

        if visualize:
            image = _draw_predictions(orig_frame, predictions, labels)
            import matplotlib.pyplot as plt

            plt.xticks([])
            plt.yticks([])
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
            plt.imshow(PIL.Image.fromarray(image))

        if return_scores:
            return predictions, labels, scores
        else:
            return predictions, labels

    def _predict_video(
        self,
        input_video_path,
        metadata_file,
        threshold=0.5,
        nms_overlap=0.1,
        track=False,
        visualize=False,
        output_file_path=None,
        multiplex=False,
        multiplex_file_path=None,
        tracker_options={
            "assignment_iou_thrd": 0.3,
            "vanish_frames": 40,
            "detect_frames": 10,
        },
        visual_options={
            "show_scores": True,
            "show_labels": True,
            "thickness": 2,
            "fontface": 0,
            "color": (255, 255, 255),
        },
        resize=False,
    ):
        """
        Runs prediction on a video and appends the output VMTI predictions in the metadata file.
        =====================   ===========================================
        **Parameter**            **Description**
        ---------------------   -------------------------------------------
        input_video_path        Required. Path to the video file to make the
                                predictions on.
        ---------------------   -------------------------------------------
        metadata_file           Required. Path to the metadata csv file where
                                the predictions will be saved in VMTI format.
        ---------------------   -------------------------------------------
        threshold               Optional float. The probability above which
                                a detection will be considered.
        ---------------------   -------------------------------------------
        nms_overlap             Optional float. The intersection over union
                                threshold with other predicted bounding
                                boxes, above which the box with the highest
                                score will be considered a true positive.
        ---------------------   -------------------------------------------
        track                   Optional bool. Set this parameter as True to
                                enable object tracking.
        ---------------------   -------------------------------------------
        visualize               Optional boolean. If True a video is saved
                                with prediction results.
        ---------------------   -------------------------------------------
        output_file_path        Optional path. Path of the final video to be saved.
                                If not supplied, video will be saved at path
                                input_video_path appended with _prediction.
        ---------------------   -------------------------------------------
        multiplex               Optional boolean. Runs Multiplex using the VMTI detections.
        ---------------------   -------------------------------------------
        multiplex_file_path     Optional path. Path of the multiplexed video to be saved.
                                By default a new file with _multiplex.MOV extension is saved
                                in the same folder.
        ---------------------   -------------------------------------------
        tracking_options        Optional dictionary. Set different parameters for
                                object tracking. assignment_iou_thrd parameter is used
                                to assign threshold for assignment of trackers,
                                vanish_frames is the number of frames the object should
                                be absent to consider it as vanished, detect_frames
                                is the number of frames an object should be detected
                                to track it.
        ---------------------   -------------------------------------------
        visual_options          Optional dictionary. Set different parameters for
                                visualization.
                                show_scores boolean, to view scores on predictions,
                                show_labels boolean, to view labels on predictions,
                                thickness integer, to set the thickness level of box,
                                fontface integer, fontface value from opencv values,
                                color tuple (B, G, R), tuple containing values between
                                0-255.
        ---------------------   -------------------------------------------
        resize                  Optional boolean. Resizes the video frames to the same size
                                (chip_size parameter in prepare_data) that the model was
                                trained on, before detecting objects.
                                Note that if resize_to parameter was used in prepare_data,
                                the video frames are resized to that size instead.
                                By default, this parameter is false and the detections are run
                                in a sliding window fashion by applying the model on cropped
                                sections of the frame (of the same size as the model was
                                trained on).
        =====================   ===========================================
        """

        VideoUtils.predict_video(
            self,
            input_video_path,
            metadata_file,
            threshold,
            nms_overlap,
            track,
            visualize,
            output_file_path,
            multiplex,
            multiplex_file_path,
            tracker_options,
            visual_options,
            resize,
        )



# Copyright 2018-2019 Open-MMLab.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# cascade_rcnn_hrnetv2p_w40_20e_coco.py, box AP=43.8
_base_ = "./_base_/models/cascade_rcnn_r50_fpn.py"
model = dict(
    pretrained="open-mmlab://msra/hrnetv2_w40",
    backbone=dict(
        _delete_=True,
        type="HRNet",
        extra=dict(
            stage1=dict(
                num_modules=1,
                num_branches=1,
                block="BOTTLENECK",
                num_blocks=(4,),
                num_channels=(64,),
            ),
            stage2=dict(
                num_modules=1,
                num_branches=2,
                block="BASIC",
                num_blocks=(4, 4),
                num_channels=(40, 80),
            ),
            stage3=dict(
                num_modules=4,
                num_branches=3,
                block="BASIC",
                num_blocks=(4, 4, 4),
                num_channels=(40, 80, 160),
            ),
            stage4=dict(
                num_modules=3,
                num_branches=4,
                block="BASIC",
                num_blocks=(4, 4, 4, 4),
                num_channels=(40, 80, 160, 320),
            ),
        ),
    ),
    neck=dict(
        _delete_=True, type="HRFPN", in_channels=[40, 80, 160, 320], out_channels=256
    ),
)

checkpoint = "http://download.openmmlab.com/mmdetection/v2.0/hrnet/cascade_rcnn_hrnetv2p_w40_20e_coco/cascade_rcnn_hrnetv2p_w40_20e_coco_20200512_161112-75e47b04.pth"



from arcgis.gis import User, GIS
from typing import Any, List


class Invitation:
    """Represents a user's invitation"""

    _url: str = None
    _gis: GIS = None
    _properties: dict = None

    def __init__(self, url: str, gis: GIS):
        self._url = url
        self._gis = gis

    @property
    def properties(self) -> dict:
        """returns the details of the invitation"""
        if self._properties is None:
            self._properties = self._gis._con.get(self._url, {"f": "json"})
        return self._properties

    def accept(self) -> dict:
        """
        When a group owner or an administrator invites a user to their
        group, it results in a user invitation. The invited user can
        accept the invitation using the `accept` operation. This
        operation adds the invited user to the group, and the invitation is
        deleted. This operation also creates a notification for the user
        indicating that the user's invitation was accepted.

        :return: Dict[str:str]
        """
        url = f"{self._url}/accept"
        res = self._gis._con.post(url, {"f": "json"})
        return res

    def decline(self) -> dict:
        """
        When a group administrator invites a user to their group, it
        results in a group invitation. The invited user can decline the
        invitation using the `decline` operation. The operation deletes the
        invitation and creates a notification for the user indicating that
        they declined the invitation. The invited user is not added to the
        group.

        :returns: Dict[str:str]
        """
        url = f"{self._url}/decline"
        res = self._gis._con.post(url, {"f": "json"})
        return res


class UserInvitationManager:
    """Provides access to the user's invitations"""

    _user = None
    _gis = None
    _url = None

    def __init__(self, user: User):
        self._user: User = user
        self._gis: GIS = user._gis

        self._url = (
            f"{self._gis._portal.resturl}community/users/{user.username}/invitations"
        )

    @property
    def list(self) -> List[Invitation]:
        """
        returns a list of all invitations

        :return: List of :class:`~arcgis.gis.sharing._users.Invitation` objects

        """
        resp = self._gis._con.get(self._url, {"f": "json"})
        return [
            Invitation(self._url + f"/{invites['id']}", self._gis)
            for invites in resp.get("userInvitations", [])
        ]



from ._base import (
    BaseCloneDefinition,
    BaseCloneItemDefinition,
    BaseCloneTextItemDefinition,
)
from ._mgr import register, unregister, clone_registry
from ._groups import GroupCloner, CloningJob
from ._ux import UXCloner

__all__ = [
    "register",
    "unregister",
    "clone_registry",
    "BaseCloneDefinition",
    "BaseCloneItemDefinition",
    "BaseCloneTextItemDefinition",
    "GroupCloner",
    "CloningJob",
    "UXCloner",
]



import asyncio
import json
import re
import uuid
from collections import namedtuple
from copy import deepcopy
from functools import lru_cache
from pathlib import Path
from typing import Union, Awaitable, Iterable, Optional
from warnings import warn

import pandas as pd

from arcgis._impl.common._utils import _lazy_property as lazy_property
from arcgis.features import GeoAccessor, FeatureSet
from arcgis.geometry import Geometry, SpatialReference
from arcgis.gis import GIS
from arcgis.network.analysis import get_travel_modes
from ._spatial import change_spatial_reference
from ._utils import (
    add_proximity_to_enrich_feature_list,
    extract_from_kwargs,
    get_helper_service_url,
    get_sanitized_names,
    get_spatially_enabled_dataframe,
    is_dict_geometry,
    is_dict_featureset,
    local_vs_gis,
    pep8ify,
    pro_at_least_version,
    run_async,
    set_source,
    validate_network_travel_mode,
)

__all__ = ["BusinessAnalyst", "Country"]


class AOI(object):
    """
    An AOI (area of interest) delineates the area being used for enrichment.
    Currently this is implemented as a Country, but is classed as a parent
    so later areas of interest can be delineated spanning country borders.
    """

    def __init__(self, source, **kwargs):
        # set the enrichment property
        self._ba = self._ba = (
            kwargs["enrichment"] if "enrichment" in kwargs else BusinessAnalyst(source)
        )

        # set the source
        self.source = self._ba.source

        # placeholder for properties
        self.properties: namedtuple = None

    def __repr__(self):
        repr_str = f"<{type(self).__name__} ({self.source})>"
        return repr_str

    @property
    def source(self) -> Union[str, GIS]:
        """
        Source being used.

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        in_source               Optional either the 'local' keyword or an instantiated ``GIS`` object
                                instance.
        ==================      ====================================================================
        """
        return self._source

    @source.setter
    def source(self, in_source: Optional[Union[str, GIS]] = None) -> None:
        self._source = set_source(in_source)

        # if working with a GIS object instance, we need to set a few extra properties
        if isinstance(self._source, GIS):
            # run a few checks and get the helper service for geoenrichment
            self._base_url = get_helper_service_url(self.source, "geoenrichment")

            # run a check for nuances of hosted notebooks
            if self._source._is_hosted_nb_home:
                # check if there is a private service url (only true for hosted notebooks)
                res = self._source._private_service_url(self._base_url)

                # if there is a private service url returned in the response, change to this
                self._base_url = (
                    res["privateServiceUrl"]
                    if "privateServiceUrl" in res
                    else res["serviceUrl"]
                )

    @lazy_property
    @local_vs_gis
    def enrich_variables(self):
        """Pandas DataFrame of available enrichment enrich_variables."""
        pass

    def _enrich_variables_gis(self):
        """GIS implementation of enrich_variables property."""
        # if the AOI has an iso3 property, then is a Country, and should be added as part of the request
        iso3 = self.__dict__["iso3"] if hasattr(self, "iso3") else None

        # get the enrich variables
        ev = self._ba._get_enrich_variables_gis(iso3)

        return ev

    def get_enrich_variables_from_iterable(
        self, enrich_variables: Union[Iterable, pd.Series], **kwargs
    ) -> pd.DataFrame:
        """Get a dataframe of enrich enrich_variables associated with the list of enrich_variables
        passed in. This is especially useful when needing aliases (*human readable
        names*), or are interested in enriching more data using previously enriched
        data as a template.

        ==================      ====================================================================
        **Parameter**            **Description**
        ------------------      --------------------------------------------------------------------
        enrich_variables        Required Iterable (normally a list) of enrich_variables correlating to
                                enrichment enrich_variables. These variable names can be simply the name, the
                                name prefixed by the collection separated by a dot, or the output from
                                enrichment in ArcGIS Pro with the field name modified to fit field naming
                                and length constraints.
        ==================      ====================================================================

        :returns:
            Pandas DataFrame of enrich enrich_variables with the different available aliases.

        .. code-block:: python

            from pathlib import Path

            import arcpy
            from business_analyst import BusinessAnalyst, Country

            # paths
            gdb = Path(r'C:/path/to/geodatabase.gdb')
            enriched_fc_pth = gdb/'enriched_data'
            features_to_be_enriched = gdb/'block_groups_pdx'
            new_fc_pth = gdb/'block_groups_pdx_enriched'

            # get a list of column names from previously enriched data
            attr_lst = [c.name for c in arcpy.ListFields(str(enriched_fc_pth))

            # get a country to work in
            cntry = BusinessAnalyst('local').get_country('USA')

            # get dataframe of enrich_variables used for previously enriched data
            enrich_vars = cntry.get_enrich_variables_from_name_list(attr_lst)

            # enrich block groups in new area of interest using the same enrich_variables
            bg_df = pd.DataFrame.spatial.from_featureclass(features_to_be_enriched)
            enrich_df = cntry.enrich(new_fc_pth)

            # save the enriched data
            enrich_df.spatial.to_featureclass(new_fc_pth)
        """
        # call the method from the parent business analyst instance
        ev = self._ba.get_enrich_variables_from_iterable(
            enrich_variables, country=self, **kwargs
        )

        return ev

    @lazy_property
    @local_vs_gis
    def geography_levels(self) -> pd.DataFrame:
        """Dataframe of available geography levels."""
        pass

    def get_geography_level(
        self,
        selector: Optional[Union[str, pd.DataFrame]] = None,
        selection_field: str = "NAME",
        query_string: Optional[str] = None,
        output_spatial_reference: Optional[Union[SpatialReference, dict, int]] = 4326,
        return_geometry: Optional[bool] = True,
    ) -> pd.DataFrame:
        """
        Get a DataFrame at an available geography_level level.

        Args:
            selector: Spatially Enabled DataFrame or string value used to select features.
                If a specific value can be identified using a string, even if
                just part of the field value, you can insert it here.
            selection_field: This is the field to be searched for the string values
                input into selector. It defaults to ``NAME``.
            query_string: If a more custom query is desired to filter the output, please
                use SQL here to specify the query. The normal query is ``UPPER(NAME) LIKE
                UPPER('%<selector>%')``. However, if a more specific query is needed, this
                can be used as the starting point to get more specific.
            output_spatial_reference: Desired spatial reference for returned data. This can be
                a spatial reference WKID as an integer, a dictionary representing the spatial
                reference, or a Spatial Reference object. The default is WGS84 (WKID 4326).
            return_geometry: Boolean indicating if geometry should be returned. While
                typically the case, there are instances where it is useful to not
                retrieve the geometries. This includes testing to create a query  only
                retrieving one area of interest.

        Returns:
            Pandas DataFrame of values fulfilling selection.
        """
        pass

    @lazy_property
    @local_vs_gis
    def travel_modes(self) -> pd.DataFrame:
        """Dataframe of available travel modes."""
        pass

    def enrich(
        self,
        geographies: Union[pd.DataFrame, Iterable, Path],
        enrich_variables: Union[pd.DataFrame, Iterable],
        return_geometry: bool = True,
        standard_geography_level: Optional[Union[int, str]] = None,
        standard_geography_id_column: Optional[str] = None,
        proximity_type: Optional[str] = None,
        proximity_value: Optional[Union[float, int]] = None,
        proximity_metric: Optional[str] = None,
        output_spatial_reference: Union[int, dict, SpatialReference] = None,
        estimate_credits: bool = False,
        **kwargs,
    ) -> Union[pd.DataFrame, Path, float]:
        """
        Get demographic data apportioned to the input geographies based on population density
        weighting. Input geographies can be polygons or points. If providing point geographies,
        an area surrounding every point will be used to determine the area for enrichment.

        .. note::

            By default, point area is typically a buffered straight-line distance. However,
            if a transportation network is available, more accurate proximity metrics may be
            available, such as drive distance or drive time. This is the case if using ``local``
            (ArcGIS Pro with Business Analyst and local data) or a ``GIS`` object connected to ArcGIS
            Online, and very well may also the be the case if using an instance of ArcGIS Enterprise.

        =============================       ====================================================================
        **Parameter**                        **Description**
        -----------------------------       --------------------------------------------------------------------
        geographies                         Required geographic areas or points to be enriched.
                                            enrich_variables: Enrichment enrich_variables to be used,
                                            typically discovered using the "enrich_variables" property.
        -----------------------------       --------------------------------------------------------------------
        return_geometry                     Optional boolean indicating if geometry is desired in the output.
                                            Default is True.
        -----------------------------       --------------------------------------------------------------------
        standard_geography_level            If the input geographies are a standard geography level,
                                            it can be specified using either the standard geography index or the standard
                                            geography identifier retrieved using the Country.geography_levels property.
        -----------------------------       --------------------------------------------------------------------
        standard_geography_id_column        Column with values uniquely identifying the input
                                            geographies using a standard level of geography. For example, in the United
                                            States, typically block groups are used for analysis if possible, and these
                                            block groups all have a unique identifier, typically referred to as the FIPS.
                                            If you have this value in a column of your data, it will *dramatically* speed
                                            up the enrichment process if you specify it in this parameter.
        -----------------------------       --------------------------------------------------------------------
        proximity_type                      Type of area to create around each point.
        -----------------------------       --------------------------------------------------------------------
        proximity_value                     Scalar value representing the proximity around each point to
                                            be used for creating an area for enrichment. For instance, if using 1.2 km,
                                            the input for this parameter is 1.2. The default is 1.
        -----------------------------       --------------------------------------------------------------------
        proximity_metric                    Scalar metric defining the proximity_value. Again, if
                                            1.2 km, the input for this metric will be kilometers. The default is
                                            ``kilometers``.
        -----------------------------       --------------------------------------------------------------------
        output_spatial_reference            Desired spatial reference for returned data. This can be
                                            a spatial reference WKID as an integer, a dictionary representing the spatial
                                            reference, or a Spatial Reference object. The default is WGS84 (WKID 4326).
        -----------------------------       --------------------------------------------------------------------
        estimate_credits                    While only useful for ArcGIS Online, enables populating the
                                            parameters just as you would for enriching using ArcGIS Online, and getting an
                                            estimate of the number of credits, which will be consumed if the enrich
                                            operation is performed. If this parameter is populated, the function will *not*
                                            perform the enrich operation or consume credits. Rather, it will *only* provide
                                            a credit consumption estimate.
        =============================       ====================================================================

        Returns:
            Pandas DataFrame, path to the output Feature Class or table, or float of predicted
            credit consumption.
        """
        enrich_df = self._ba.enrich(
            geographies=geographies,
            enrich_variables=enrich_variables,
            country=self,
            standard_geography_level=standard_geography_level,
            standard_geography_id_column=standard_geography_id_column,
            proximity_type=proximity_type,
            proximity_value=proximity_value,
            proximity_metric=proximity_metric,
            return_geometry=return_geometry,
            output_spatial_reference=output_spatial_reference,
            estimate_credits=estimate_credits,
            **kwargs,
        )
        return enrich_df


class Country(AOI):
    """
    Country enables access to Business Analyst functionality. Business Analyst
    data is available by iso3 using both ``local`` (ArcGIS Pro with the Business
    Analyst extension and local data) and ``GIS`` sources.

    =============================       ====================================================================
    **Parameter**                        **Description**
    -----------------------------       --------------------------------------------------------------------
    iso3                                The country's ISO3 identifier.
    -----------------------------       --------------------------------------------------------------------
    source                              Optional ``GIS`` object or ``local`` keyword specifying the Business
                                        Analyst data and analysis source. If ``local``, the Python
                                        environment *must* have ``arcpy`` installed with bindings to ArcGIS
                                        Pro with the Business Analyst extension. If connecting to a ``GIS``
                                        instance, both ArcGIS Enterprise with Business Analyst and ArcGIS
                                        Online are supported. However, please be aware, any geoenrichment or
                                        analysis *will* consume ArcGIS Online credits.
    -----------------------------       --------------------------------------------------------------------
    year                                Optional integer explicitly specifying the year to reference.
                                        This is only honored if using local resources and the specified
                                        year is available.
    =============================       ====================================================================

    """

    def __init__(
        self,
        iso3: str,
        source: Optional[Union[str, GIS]] = None,
        year: Optional[int] = None,
        **kwargs,
    ) -> None:
        # invoke the AOI init method - mostly just sets the _ba property
        super().__init__(source, **kwargs)

        # set the iso3 property based on the iso3
        self.iso3 = self._ba._standardize_country_str(iso3)

        # use the iso3 to filter the available countries to a dataframe of just the country requested.
        sel_df = self._ba.countries[self._ba.countries["iso3"] == self.iso3]

        # if the data source is local, but no year was provided, get the year
        if self.source == "local" and year is None:
            year = sel_df.vintage.max()

        # if the year is provided, validate
        elif self.source == "local" and year is not None:
            # just in case it was accidentally input as a string
            year = int(year) if isinstance(year, str) else year

            # get list of valid values for this country
            lcl_yr_vals = list(
                self._ba.countries[self._ba.countries["iso3"] == self.iso3]["vintage"]
            )

            # validate year is available for given country
            assert year in lcl_yr_vals, (
                f'The year you provided, "{year}" is not among the available '
                f'years ({", ".join([str(v) for v in lcl_yr_vals])}) for "{self.iso3.upper()}".'
            )

        # if source is GIS and year is provided, warn will be ignored
        elif isinstance(self.source, GIS) and year is not None:
            warn(
                "Explicitly specifying a year (vintage) is not supported when using a GIS instance as the source."
            )

        # get the first row from the selected dataframe, filtering based on year if local, for the properties
        properties = (
            sel_df[sel_df.vintage == year].iloc[0]
            if self.source == "local"
            else sel_df.iloc[0]
        )

        # if local, set a few more properties unique to local
        if self.source == "local":
            # to avoid confusion, add year as alias to vintage
            properties["year"] = properties["vintage"]

            # set the path to the network dataset
            properties["network_path"] = self._get_network_dataset_path_local(
                properties
            )

            # ensure analysis uses correct data source
            self._set_local_ba_analysis_source(properties)

        # convert the properties to a namedtuple and set as property of object
        self.properties = namedtuple("properties", properties.index)(*properties)

    def __repr__(self):
        repr_str = f"<{type(self).__name__} - {self.iso3}"

        if self.source == "local":
            repr_str = f"{repr_str} {self.properties.vintage} (local)>"
        else:
            repr_str = f"{repr_str} ({self.source})>"

        return repr_str

    @staticmethod
    def _get_network_dataset_path_local(properties) -> str:
        """Get the path to the network dataset for this country."""
        import arcpy

        # get a dictionary of dataset properties
        ds_dict = {ds["id"]: ds for ds in arcpy._ba.getLocalDatasets()}

        # get the path to the network dataset using the country id (captures both country and year)
        src_pth = ds_dict[properties.country_id]["networkDatasetPath"]

        return src_pth

    @staticmethod
    def _set_local_ba_analysis_source(properties) -> None:
        """Ensure analysis uses the correct data source."""
        import arcpy

        arcpy.env.baDataSource = f"LOCAL;;{properties.country_id}"
        arcpy.env.baNetworkSource = properties.network_path

    def _enrich_variables_local(self) -> pd.DataFrame:
        """Local implementation of enrich_variables property and only available by iso3 currently."""
        # ensure using correct data sources
        self._set_local_ba_analysis_source(self.properties)

        # lazy load
        from arcpy._ba import ListVariables

        # pull out the iso3 dataframe to a variable for clarity
        cntry_df = self._ba.countries

        # create a filter to select the right iso3 and year dataset
        dataset_fltr = (cntry_df["iso3"] == self.iso3) & (
            cntry_df["vintage"] == int(self.properties.vintage)
        )

        # get the iso3 identifier needed for listing enrich_variables
        ba_id = cntry_df[dataset_fltr].iloc[0]["country_id"]

        # retrieve variable objects
        var_gen = ListVariables(ba_id)

        # use a list comprehension to unpack the properties of the enrich_variables into a dataframe
        var_df = pd.DataFrame(
            (
                (v.Name, v.Alias, v.DataCollectionID, v.FullName, v.OutputFieldName)
                for v in var_gen
            ),
            columns=[
                "name",
                "alias",
                "data_collection",
                "enrich_name",
                "enrich_field_name",
            ],
        )

        return var_df

    def _geography_levels_local(self):
        """Local implementation of geography_levels."""
        import arcpy

        # ensure using correct data sources
        self._set_local_ba_analysis_source(self.properties)

        # get a dataframe of level properties for the country
        geo_lvl_df = pd.DataFrame.from_records(
            [
                lvl.__dict__
                for lvl in arcpy._ba.ListGeographyLevels(self.properties.country_id)
            ]
        )

        # reverse sorting order so smallest is at the top
        geo_lvl_df = geo_lvl_df.iloc[::-1].reset_index(drop=True)

        # calculate a field for use in the accessor
        geo_lvl_df.insert(
            0,
            "level_name",
            geo_lvl_df["LevelID"].apply(lambda val: pep8ify(val.split(".")[1])),
        )

        # purge columns and include AdminLevel if present (added in Pro 2.9)
        out_col_lst = [
            "level_name",
            "SingularName",
            "PluralName",
            "LevelName",
            "LayerID",
            "IDField",
            "NameField",
        ]
        if "AdminLevel" in list(geo_lvl_df.columns):
            out_col_lst = out_col_lst + ["AdminLevel"]
        geo_lvl_df.drop(
            columns=[c for c in geo_lvl_df.columns if c not in out_col_lst],
            inplace=True,
        )

        # rename fields for consistency
        out_rename_dict = {
            "SingularName": "singular_name",
            "PluralName": "plural_name",
            "LevelName": "alias",
            "LayerID": "level_id",
            "IDField": "id_field",
            "NameField": "name_field",
            "AdminLevel": "admin_level",
        }
        geo_lvl_df.rename(columns=out_rename_dict, inplace=True, errors="ignore")

        return geo_lvl_df

    def _geography_levels_gis(self):
        """GIS implementation of geography levels."""
        # unpack the geoenrichment url from the properties
        enrich_url = self.source.properties.helperServices.geoenrichment.url

        # construct the url to the standard geography levels
        url = f"{enrich_url}/Geoenrichment/standardgeographylevels"

        # get the geography levels from the enrichment server
        res_json = self.source._con.post(url, {"f": "json"})

        # unpack the geography levels from the json
        geog_lvls = res_json["geographyLevels"]

        # get matching geography levels out of the list of countries
        for lvl in geog_lvls:
            if lvl["countryID"] == self.properties.iso2:
                geog = lvl
                break  # once found, bust out of the loop

        # get the hierarchical levels out as a dataframe
        geo_lvl_df = pd.DataFrame(geog["hierarchies"][0]["levels"])

        # reverse the sorting so the smallest is at the top
        geo_lvl_df = geo_lvl_df.iloc[::-1].reset_index(drop=True)

        # create consistent naming convention
        geo_lvl_df["level_name"] = geo_lvl_df["id"].apply(
            lambda val: pep8ify(val.split(".")[1])
        )

        # clean up the field names so they follow more pythonic conventions and are consistent
        geo_lvl_df = geo_lvl_df[
            ["level_name", "singularName", "pluralName", "name", "id", "adminLevel"]
        ].copy()
        geo_lvl_df.columns = [
            "level_name",
            "singular_name",
            "plural_name",
            "alias",
            "level_id",
            "admin_level",
        ]

        return geo_lvl_df

    def _travel_modes_local(self):
        """Local implementation of travel modes."""
        # ensure using correct data sources
        self._set_local_ba_analysis_source(self.properties)

        # get a list of useful information about each travel mode
        mode_lst = [
            [
                mode.name,
                mode.description,
                mode.type,
                mode.impedance,
                mode.timeAttributeName,
                mode.distanceAttributeName,
            ]
            for mode in self._travel_modes_dict_local.values()
        ]

        # create a dataframe of the travel modes
        mode_df = pd.DataFrame(
            mode_lst,
            columns=[
                "alias",
                "description",
                "type",
                "impedance",
                "time_attribute_name",
                "distance_attribute_name",
            ],
        )

        # add a pep8tified name for each travel mode
        mode_name = mode_df["alias"].apply(lambda val: pep8ify(val))
        mode_df.insert(0, "name", mode_name)

        # calculate impedance category
        imped_cat = mode_df["impedance"].apply(
            lambda val: "temporal" if val.lower().endswith("time") else "distance"
        )
        insert_idx = mode_df.columns.get_loc("impedance") + 1
        mode_df.insert(insert_idx, "impedance_category", imped_cat)

        return mode_df

    @lazy_property
    def _travel_modes_dict_local(self) -> dict:
        """Since looking up the travel modes is time consuming, use a lazy property to cache it."""
        import arcpy

        return arcpy.nax.GetTravelModes(self.properties.network_path)

    def _travel_modes_gis(self):
        """Web GIS implementation of travel modes."""
        # get the travel modes FeatureSet
        trvl_df = self._ba.travel_modes

        return trvl_df

    def _enrich_local(
        self,
        geographies: Union[pd.DataFrame, Iterable, Path],
        enrich_variables: Union[pd.DataFrame, Iterable],
        return_geometry: bool = True,
        standard_geography_level: Optional[Union[int, str]] = None,
        standard_geography_id_column: Optional[str] = None,
        proximity_type: Optional[str] = None,
        proximity_value: Optional[Union[float, int]] = None,
        proximity_metric: Optional[str] = None,
        output_spatial_reference: Union[int, dict, SpatialReference] = 4326,
        estimate_credits: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Local enrich method implementation."""
        # ensure using correct data sources
        self._set_local_ba_analysis_source(self.properties)

        enrich_res = self._ba.enrich(
            geographies=geographies,
            enrich_variables=enrich_variables,
            country=self,
            standard_geography_level=standard_geography_level,
            standard_geography_id_column=standard_geography_id_column,
            proximity_type=proximity_type,
            proximity_value=proximity_value,
            proximity_metric=proximity_metric,
            return_geometry=return_geometry,
            output_spatial_reference=output_spatial_reference,
            estimate_credits=estimate_credits,
        )
        return enrich_res


class BusinessAnalyst(object):
    """
    The BusinessAnalyst object enables access to Business Analyst functionality thorough a specified
    source. A source can either be in an environment with ArcGIS Pro with Business Analyst and local
    data (``local``) or a Web GIS (``GIS``) object instance. The Web GIS can reference either an
    instance of ArcGIS Enterprise with Business Analyst or ArcGIS Online.

    .. note::

        If the source is not explicitly set, the ``BusinessAnalyst`` object will first try to use
        ``local``, ArcGIS Pro with the Business Analyst extension. If this is not available,
        ``BusinessAnalyst`` will try to use a ``GIS`` object instance already in the session. If
        neither is available, and a ``source`` is not set, this will invoke an error.

    .. warning::

        GeoEnrichment (adding demographic enrich_variables) using ArcGIS Online *does* cost credits.
        Country (``BusinessAnalyst.countries``) and variable (:func:`~arcgis.geoenrichment.Country.enrich_variables`)
        introspection does *not* cost any credits.

    =============================       ====================================================================
    **Parameter**                        **Description**
    -----------------------------       --------------------------------------------------------------------
    source                              Optional ``GIS`` object or ``local`` keyword specifying the Business
                                        Analyst data and analysis source. If ``local``, the Python
                                        environment *must* have ``arcpy`` installed with bindings to ArcGIS
                                        Pro with the Business Analyst extension. If connecting to a ``GIS``
                                        instance, both ArcGIS Enterprise with Business Analyst and ArcGIS
                                        Online are supported. However, please be aware, any geoenrichment or
                                        analysis *will* consume ArcGIS Online credits.
    =============================       ====================================================================

    """

    def __init__(self, source: Optional[Union[str, GIS]] = None) -> None:
        # set the source, defaulting, based on what is available, to local or active_gis, or simply error if neither
        self.source = source

    def __repr__(self):
        repr_str = f"<{type(self).__name__} ({self.source})>"
        return repr_str

    @property
    def source(self) -> Union[str, GIS]:
        """
        Source being used.

        Args:
            in_source: Either the 'local' keyword or an instantiated ``GIS`` object
                instance.
        """
        return self._source

    @source.setter
    def source(self, in_source: Optional[Union[str, GIS]] = None) -> None:
        self._source = set_source(in_source)

        # if working with a GIS object instance, we need to set a few extra properties
        if isinstance(self._source, GIS):
            # run a few checks and get the helper service for geoenrichment
            self._base_url = get_helper_service_url(self.source, "geoenrichment")

            # run a check for nuances of hosted notebooks
            if self._source._is_hosted_nb_home:
                # check if there is a private service url (only true for hosted notebooks)
                res = self._source._private_service_url(self._base_url)

                # if there is a private service url returned in the response, change the url being used to this
                self._base_url = (
                    res["privateServiceUrl"]
                    if "privateServiceUrl" in res
                    else res["serviceUrl"]
                )

    @lazy_property
    @local_vs_gis
    def countries(self) -> pd.DataFrame:
        """DataFrame of available countries with relevant metadata columns based on the source."""
        pass

    def _countries_local(self) -> pd.DataFrame:
        """Local countries implementation."""
        # lazy load to avoid import issues
        import arcpy._ba

        # get a generator of dataset objects
        ds_lst = list(arcpy._ba.ListDatasets())

        # throw error if no local datasets are available
        assert len(ds_lst), (
            "No datasets are available locally. If you want to locate available countries on a "
            "Web GIS, please provide a GIS object instance for the source parameter when creating "
            "the BusinessAnalyst object."
        )

        # organize all the iso3 dataset properties
        cntry_lst = [
            (
                ds.CountryInfo.Name,
                ds.Version,
                ds.CountryInfo.ISO2,
                ds.CountryInfo.ISO3,
                ds.DataSourceID,
                ds.ID,
                None,
            )
            for ds in ds_lst
        ]

        # create a dataframe of the iso3 properties
        cntry_df = pd.DataFrame(
            cntry_lst,
            columns=[
                "name",
                "vintage",
                "iso2",
                "iso3",
                "data_source_id",
                "country_id",
                "hierarchies",
            ],
        )

        # convert the vintage years to integer
        cntry_df["vintage"] = cntry_df["vintage"].astype("int64")

        # ensure the values are in order by iso3 and year
        cntry_df.sort_values(["iso3", "vintage"], inplace=True)

        # organize the columns
        cntry_df = cntry_df[
            [
                "iso2",
                "iso3",
                "name",
                "vintage",
                "country_id",
                "data_source_id",
                "hierarchies",
            ]
        ]

        return cntry_df

    def _countries_gis(self) -> pd.DataFrame:
        """GIS countries implementation."""
        # make sure countries are available
        ge_err_msg = (
            "The provided GIS instance does not appear to have geoenrichment enabled and configured, "
            "so no countries are available."
        )
        assert "geoenrichment" in self.source.properties.helperServices, ge_err_msg
        assert isinstance(
            self.source.properties.helperServices.geoenrichment["url"], str
        ), ge_err_msg

        # extract out the geoenrichment url
        ge_url = self.source.properties.helperServices.geoenrichment["url"]
        if self.source._is_hosted_nb_home:
            res = self.source._private_service_url(ge_url)
            ge_url = (
                res["privateServiceUrl"]
                if "privateServiceUrl" in res
                else res["serviceUrl"]
            )

        # get a list of countries available on the Web GIS for enrichment
        url = f"{ge_url}/Geoenrichment/Countries"
        cntry_res = self.source._con.post(url, {"f": "json"})
        cntry_dict = cntry_res["countries"]

        # convert the dictionary to a dataframe
        cntry_df = pd.DataFrame(cntry_dict)

        # clean up some column names for consistency
        cntry_df.rename(
            {
                "id": "iso2",
                "abbr3": "iso3",
                "altName": "alt_name",
                "defaultDatasetID": "default_dataset",
                "hierarchies": "hierarchy",
            },
            inplace=True,
            axis=1,
        )
        keep_cols = [
            "iso2",
            "iso3",
            "name",
            "alt_name",
            "datasets",
            "default_dataset",
            "continent",
            "hierarchy",
        ]
        cntry_df = cntry_df[keep_cols]

        # clean up column for hierarchies to only keep alias if simple
        alias_names = []
        for i, v in cntry_df["hierarchy"].items():
            cntry_hier = []
            for hier in v:
                cntry_hier.append(hier["ID"])
            alias_names.append(cntry_hier)
        cntry_df["hierarchy"] = alias_names

        return cntry_df

    def _get_hierarchies_df(self, country_string: str):
        """Internal helper method to get the dataframe of hierarchies for each country"""
        # make sure countries are available
        ge_err_msg = (
            "The provided GIS instance does not appear to have geoenrichment enabled and configured, "
            "so no countries are available."
        )
        assert "geoenrichment" in self.source.properties.helperServices, ge_err_msg
        assert isinstance(
            self.source.properties.helperServices.geoenrichment["url"], str
        ), ge_err_msg

        # extract out the geoenrichment url
        ge_url = self.source.properties.helperServices.geoenrichment["url"]
        if self.source._is_hosted_nb_home:
            res = self.source._private_service_url(ge_url)
            ge_url = (
                res["privateServiceUrl"]
                if "privateServiceUrl" in res
                else res["serviceUrl"]
            )

        # get a list of countries available on the Web GIS for enrichment
        url = f"{ge_url}/Geoenrichment/Countries"
        cntry_res = self.source._con.post(url, {"f": "json"})
        cntry_dict = cntry_res["countries"]

        # convert the dictionary to a dataframe
        cntry_df = pd.DataFrame(cntry_dict)

        # clean up some column names for consistency
        cntry_df.rename(
            {
                "abbr3": "iso3",
            },
            inplace=True,
            axis=1,
        )
        keep_cols = ["iso3", "hierarchies"]
        cntry_df = cntry_df[keep_cols]

        # Get dataframe for specific country we are working with
        cntry_interest_df = cntry_df[cntry_df["iso3"] == country_string]
        # Get only the hierarchy column value and create own dataframe from it
        hierarchy_df = pd.DataFrame(cntry_interest_df.iloc[0]["hierarchies"])

        keep_cols = [
            "ID",
            "alias",
            "shortDescription",
            "datasets",
            "levelsInfo",
            "variablesInfo",
            "hasInterestingFactsStatistics",
        ]
        hierarchy_df = hierarchy_df[keep_cols]
        return hierarchy_df

    def _standardize_country_str(self, country_string: str) -> str:
        """Internal helper method to standardize the input for iso3 identifier strings to ISO3."""
        # cast to lowercase to sidestep any issues with case
        cntry_str = country_string.lower()

        # filter functions for getting the iso3 iso3 value
        iso3_fltr = self.countries["iso3"].str.lower() == cntry_str
        iso2_fltr = self.countries["iso2"].str.lower() == cntry_str
        name_fltr = self.countries["name"].str.lower() == cntry_str

        # construct the filter, using alias if working online
        cntry_fltr = iso3_fltr | iso2_fltr | name_fltr
        if isinstance(self.source, GIS):
            alias_fltr = self.countries["alt_name"].str.lower() == cntry_str
            cntry_fltr = cntry_fltr | alias_fltr

        # query available countries to see if the requested iso3 is available
        fltr_df = self.countries[cntry_fltr]

        if len(fltr_df.index) > 0:
            iso3_str = fltr_df.iloc[0]["iso3"]
        else:
            raise ValueError(
                f'The provided iso3 code, "{country_string}" does not appear to be available. Please '
                f"choose from the available country iso3 codes discovered using the "
                f"BusinessAnalyst.countries property."
            )

        return iso3_str

    def get_country(self, iso3: str, year: Optional[int] = None) -> Country:
        """
        Get a Country object instance.
        =============================       ====================================================================
        **Parameter**                        **Description**
        -----------------------------       --------------------------------------------------------------------
        iso3                                Required String. The country's ISO3 identifier.
        -----------------------------       --------------------------------------------------------------------
        year                                Optional integer explicitly specifying the year to reference.
                                            This is only honored if using local resources and the specified
                                            year is available.
        =============================       ====================================================================

        Returns:
            Country object instance.

        A Country object instance can be created to work with Business Analyst data installed
        either locally with ArcGIS Pro or remotely through a Web GIS using a GIS object instance.

        .. code-block:: python

            from business_analyst import BusinessAnalyst
            ba = BusinessAnalyst('local')  # using ArcGIS Pro with Business Analyst
            usa = ba.get_country('USA')  # USA with most current data installed on machine

        .. code-block:: python

            from business_analyst import BusinessAnalyst
            ba = BusinessAnalyst('local')
            usa = ba.get_country('USA', year=2019)  # explicitly specifying the year with local data

        .. code-block:: python

            from arcgis.gis import GIS
            from business_analyst import BusinessAnalyst
            gis = GIS(username='my_username', password='$up3r$3cr3tP@$$w0rd')
            ba = BusinessAnalyst(gis)  # using ArcGIS Online
            usa = ba.get_country('USA')

        Once instantiated, available enrichment enrich_variables can be discovered and used for
        geoenrichment.

        .. code-block:: python

            # get the enrichment enrich_variables as a Pandas DataFrame
            evars = usa.enrich_variables

            # filter to just current year key enrich_variables
            kvars = [
                (evars.name.str.endswith('CY'))
                & (evars.data_collection.str.startswith('Key')
            ]

        Then, based on the environment being used for GeoEnrichment, these variable identifiers
        can be formatted for input into the respective enrich functions, `EnrichLayer`_ in Pro
        or `enrich`_ in the Python API. In the code below, the local and the Web GIS blocks will
        produce very similar outputs with the only differences being the column names.

        .. code-block:: python

            from arcgis.features import GeoAccessor
            import pandas as pd

            input_fc = 'C:/path/to/data.gdb/feature_class'
            output_pth = 'C:/path/to/data.gdb/enrich_features'

            # local geoenrichment
            import arcpy
            kvars_lcl = ';'.join(kvars.enrich_name)  # combine into semicolon separated string
            enrich_pth = arcpy.ba.EnrichLayer(input_fc, out_feature_class=output_pth, enrich_variables=kvars_lcl)[0]
            enrich_df = pd.DataFrame.spatial.from_featureclass(enrich_pth)  # dataframe for more analysis

            # WebGIS geoenrichment
            from arcgis.geoenrichment import enrich
            kvars_gis = ';'.join(kvars.name)
            in_df = pd.DataFrame.spatial.from_featureclass(input_fc)
            enrich_df = enrich(in_df, analysis_variables=kvars_gis, return_geometry=False)
            enrich_pth = enrich_df.spatial.to_table(output_pth)  # saving so do not have to re-run

        .. _EnrichLayer: https://pro.arcgis.com/en/pro-app/latest/tool-reference/business-analyst/enrich-layer-advanced.htm
        .. _enrich: https://developers.arcgis.com/python/api-reference/arcgis.geoenrichment.html#enrich

        """
        # standardize iso3 string to iso3 identifier
        iso3 = self._standardize_country_str(iso3)

        # create a iso3 object instance
        cntry = Country(iso3, year=year, enrichment=self)

        return cntry

    @lazy_property
    @local_vs_gis
    def enrich_variables(self) -> pd.DataFrame:
        """Available enrichment variables."""
        pass

    def _enrich_variables_gis(self) -> pd.DataFrame:
        """Web GIS implementation of enrich_variables property"""
        # retrieve the enrich variables using the method
        ev = self._get_enrich_variables_gis()

        return ev

    @lru_cache(maxsize=255)
    def _get_enrich_variables_gis(self, iso3: Optional[str] = None) -> pd.DataFrame:
        """Provide method to return enrich variables at both the BusinessAnalyst and AOI (Country) levels."""
        # construct the url with the option to simply not explicitly specify a iso3
        url = f"{self._base_url}/Geoenrichment/DataCollections/"
        if iso3 is not None:
            url = f"{url}{iso3}"

        # get the data collections from the GIS enrichment REST endpoint
        res = self.source._con.get(url, params={"f": "json"})
        msg_dc = (
            "Could not retrieve enrichment enrich_variables (DataCollections) from "
            "the GIS instance."
        )
        assert "DataCollections" in res.keys(), msg_dc

        # list to store all the dataframes as they are created for each data collection
        mstr_lst = []

        # iterate the data collections
        for col in res["DataCollections"]:
            # create a dataframe of the enrich_variables, keep only needed columns, and add the data collection name
            coll_df = pd.json_normalize(col["data"])[
                ["id", "alias", "description", "vintage", "units"]
            ]
            coll_df["data_collection"] = col["dataCollectionID"]

            # schema cleanup
            coll_df.rename(columns={"id": "name"}, inplace=True)
            coll_df = coll_df[
                ["name", "alias", "data_collection", "description", "vintage", "units"]
            ]

            # append the list
            mstr_lst.append(coll_df)

        # combine all the dataframes into a single master dataframe
        var_df = pd.concat(mstr_lst)

        # create the column for enrichment
        var_df.insert(3, "enrich_name", var_df.data_collection + "." + var_df.name)

        # create column for matching to previously enriched column names
        regex = re.compile(r"(^[0-9]+)")
        fld_vals = var_df.enrich_name.apply(
            lambda val: regex.sub(r"F\1", val.replace(".", "_"))
        )
        var_df.insert(4, "enrich_field_name", fld_vals)

        # reset the index so it looks like a single dataframe
        var_df.reset_index(inplace=True, drop=True)

        return var_df

    def get_enrich_variables_from_iterable(
        self, enrich_variables: Union[Iterable, pd.Series], **kwargs
    ) -> pd.DataFrame:
        """
        Get a dataframe of enrich enrich_variables associated with the list of enrich_variables
        passed in. This is especially useful when needing aliases (*human readable
        names*), or are interested in enriching more data using previously enriched
        data as a template.

        =============================       ====================================================================
        **Parameter**                        **Description**
        -----------------------------       --------------------------------------------------------------------
        enrich_variables                    Iterable (normally a list) of enrich_variables correlating to
                                            enrichment enrich_variables. These variable names can be simply the name, the
                                            name prefixed by the collection separated by a dot, or the output from
                                            enrichment in ArcGIS Pro with the field name modified to fit field naming
                                            and length constraints.
        =============================       ====================================================================

        Returns:
            Pandas DataFrame of enrich enrich_variables with the different available aliases.

        .. code-block:: python

            from pathlib import Path

            import arcpy
            from arcgis.apps.ba import BusinessAnalyst

            # paths
            gdb = Path(r'C:/path/to/geodatabase.gdb')
            enriched_fc_pth = gdb/'enriched_data'
            features_to_be_enriched = gdb/'block_groups_pdx'
            new_fc_pth = gdb/'block_groups_pdx_enriched'

            # ArcGIS Online credentials
            usr = 'username'
            pswd = 'P@$$w0rd!'

            # create a business analyst instance
            gis = GIS(username=usr, password=pswd)
            ba = BusinessAnalyst(gis)

            # get a list of column names from previously enriched data
            attr_lst = [c.name for c in arcpy.ListFields(str(enriched_fc_pth))

            # get dataframe of enrich_variables used for previously enriched data
            enrich_vars = cntry.get_enrich_variables_from_name_list(attr_lst)

            # enrich block groups in new area of interest using the same enrich_variables
            bg_df = pd.DataFrame.spatial.from_featureclass(features_to_be_enriched)
            enrich_df = cntry.enrich(bg_df)

            # save the enriched data
            enrich_df.spatial.to_featureclass(new_fc_pth)
        """
        # validate the series input
        ev_msg = "Only an list or list-like object can be used as input as the enrich_variables input."
        ev_valid = not isinstance(enrich_variables, pd.DataFrame) and isinstance(
            enrich_variables, (Iterable, pd.Series)
        )
        assert ev_valid, ev_msg

        # if a country is included in the kwargs, pluck it out and get the right ba source
        if "country" in kwargs.keys():
            cntry: Country = kwargs["country"]
            iso3 = cntry.iso3
            ba = cntry._ba
        else:
            cntry, iso3 = None, None
            ba = self

        # based on the source, get the available enrichment variables
        if isinstance(ba.source, GIS):
            ev = ba._get_enrich_variables_gis(iso3)
        else:
            ev = cntry.enrich_variables

        # if the input is not a series, make it so
        if isinstance(enrich_variables, Iterable):
            enrich_variables = pd.Series(enrich_variables)

        # make the enrich variables lowercase to address potential missed matches due simply to case
        enrich_variables = enrich_variables.str.lower()

        # columns from enrich variables dataframe to search for matches in
        match_cols = ["enrich_name", "enrich_field_name", "name", "alias"]

        # iteratively check for matches and if matches found, bingo out
        for col in match_cols:
            sel_vars = (
                ev[ev[col].str.lower().isin(enrich_variables)]
                .drop_duplicates(col)
                .reset_index(drop=True)
            )
            if len(sel_vars):
                break

        # if no matches found, try sanitizing the names and see if this works
        if len(sel_vars) == 0:
            enrich_variables = get_sanitized_names(enrich_variables)
            for col in match_cols:
                sel_vars = (
                    ev[ev[col].str.lower().isin(enrich_variables)]
                    .drop_duplicates(col)
                    .reset_index(drop=True)
                )
                if len(sel_vars):
                    break

        # make sure something was found, but don't break the runtime
        if "suppress_warn" not in kwargs.keys() and len(sel_vars) == 0:
            warn(f"It appears none of the input enrich enrich_variables were found.")

        return sel_vars

    def _enrich_variable_preprocessing(
        self, enrich_variables: Union[Iterable, pd.DataFrame], **kwargs
    ) -> str:
        """Provide flexibility for enrich variable preprocessing by enabling enrich enrich_variables
        to be specified in a variety of iterables, but always provide a standardized variable
        DataFrame as output.

        =============================       ====================================================================
        **Parameter**                        **Description**
        -----------------------------       --------------------------------------------------------------------
        enrich_variables                    Iterable (normally a list) or pd.DataFrame
                                            of enrich_variables correlating to
                                            enrichment enrich_variables. These variable names can be simply the name, the
                                            name prefixed by the collection separated by a dot, or the output from
                                            enrichment in ArcGIS Pro with the field name modified to fit field naming
                                            and length constraints.
        =============================       ====================================================================

        Returns:
            Pandas DataFrame of enrich enrich_variables.
        """
        # enrich variable dataframe column name
        enrich_str_col = "enrich_name"
        enrich_nm_col = "name"

        # if just a single variable is provided pipe it into a pandas series
        if isinstance(enrich_variables, str):
            enrich_variables = pd.Series([enrich_variables])

        # toss the enrich_variables into a pandas Series if an iterable was passed in
        elif isinstance(enrich_variables, Iterable) and not isinstance(
            enrich_variables, pd.DataFrame
        ):
            enrich_variables = pd.Series(enrich_variables)

        # if the enrich dataframe is passed in, check to make sure it has what we need, the right columns
        if isinstance(enrich_variables, pd.DataFrame):
            assert enrich_str_col in enrich_variables.columns, (
                f"It appears the dataframe used for enrichment does"
                f" not have the column with enrich string names "
                f"({enrich_str_col})."
            )
            assert enrich_nm_col in enrich_variables.columns, (
                f"It appears the dataframe used for enrichment does "
                f"not have the column with the enrich enrich_variables names "
                f"({enrich_nm_col})."
            )
            enrich_vars_df = enrich_variables

        # otherwise, create an enrich enrich_variables dataframe from the enrich series for a few more checks
        else:
            # get the enrich enrich_variables dataframe
            enrich_vars_df = self.get_enrich_variables_from_iterable(
                enrich_variables, **kwargs
            )

        # now, drop any duplicates so we're not getting the same variable twice from different data collections
        enrich_vars_df = enrich_vars_df.drop_duplicates("name").reset_index(drop=True)

        # note any enrich_variables submitted, but not found
        if len(enrich_variables) > len(enrich_vars_df.index):
            missing_count = len(enrich_variables) - len(enrich_vars_df.index)
            warn(
                "Some of the enrich_variables provided are not available for enrichment "
                f"(missing count: {missing_count:,}).",
                UserWarning,
            )

        # check to make sure there are enrich_variables for enrichment
        if len(enrich_vars_df.index) == 0:
            raise Exception(
                "There appear to be no enrich_variables selected for enrichment."
            )

        # get the string names needed in a list
        ev_lst = list(enrich_vars_df[enrich_str_col])

        # format output based on enrichment source
        enrich_vars = (
            json.dumps(ev_lst) if isinstance(self.source, GIS) else ";".join(ev_lst)
        )

        return enrich_vars

    @lazy_property
    @local_vs_gis
    def travel_modes(self) -> pd.DataFrame:
        """Dataframe of available travel modes."""
        pass

    def _travel_modes_gis(self):
        """Web GIS implementation of travel modes."""
        # get the travel modes FeatureSet
        trvl_fs = get_travel_modes(self.source).supported_travel_modes

        # pluck out the travel modes metadata from the TravelMode attribute in the FeatureSet
        trvl_df = pd.DataFrame.from_records(
            [json.loads(itm.attributes["TravelMode"]) for itm in trvl_fs.features]
        )

        # put the travel mode dict as a string (less the description) into a column - useful when needing for REST calls
        trvl_df["travel_mode_dict"] = [
            json.dumps(val)
            for val in trvl_df.drop(columns="description").to_dict("records")
        ]

        # make the column names pep8 compliant since in happy Python land
        trvl_df.columns = [pep8ify(c) for c in trvl_df.columns]

        # move the name into the alias and pep8ify the name
        trvl_df["alias"] = trvl_df["name"]
        trvl_df["name"] = trvl_df["name"].apply(lambda val: pep8ify(val))

        # rename a few columns for the sake of consistency with results returned from local network introspection
        trvl_df.rename(
            columns={"impedance_attribute_name": "impedance", "id": "travel_mode_id"},
            inplace=True,
        )

        # calculate impedance categories for ease of filtering in some workflows
        trvl_df["impedance_category"] = trvl_df["impedance"].apply(
            lambda val: ("temporal" if val.endswith("Time") else "distance")
            if pd.notna(val)
            else val
        )

        # reorganize the column order
        trvl_df = trvl_df[
            [
                "name",
                "alias",
                "description",
                "type",
                "impedance",
                "impedance_category",
                "time_attribute_name",
                "distance_attribute_name",
                "travel_mode_id",
                "travel_mode_dict",
            ]
        ]

        return trvl_df

    def _can_use_arrow(self, geo) -> bool:
        if not isinstance(geo, pd.DataFrame):
            return False  # not a DataFrame

        if not geo.spatial.validate():
            return False  # not a valid SeDF

        if len(geo.spatial.geometry_type) != 1:
            return False  # more than one geometry type

        if geo.spatial.geometry_type[0] != "polygon":
            return False  # only polygonal DF are supported at this point

        return True

    def _enrich_using_arrow(self, in_sedf, variables) -> pd.DataFrame:
        # create a data frame with two columns: object id and shape in WKB
        input_shape_series = in_sedf.loc[:, in_sedf.spatial.name]
        df_input = pd.DataFrame({in_sedf.spatial.name: input_shape_series})
        df_input.spatial.set_geometry(in_sedf.spatial.name)

        # come up with index field that doesn't exist yet
        orig_index_field = str(uuid.uuid4())
        # enrichArrowTable returns records in arbitrary order, ORIG_INDEX is in the original order
        # ORIG_INDEX starts with 0
        df_input[orig_index_field] = range(0, len(in_sedf))

        geo_accessor = GeoAccessor(df_input)
        arrow_table = geo_accessor.to_arrow()

        import arcpy._ba

        output_table = arcpy._ba.enrichArrowTable(arrow_table, variables, False)

        enrich_result_df = output_table.to_pandas()

        input_copy_df = in_sedf.copy()
        input_copy_df[orig_index_field] = range(0, len(in_sedf))

        # we need to rename that field to avoid clashes
        enrich_result_df.rename(columns={"ORIG_INDEX": orig_index_field}, inplace=True)

        if "ORIG_OID" in enrich_result_df:
            enrich_result_df.drop(["ORIG_OID"], axis=1, inplace=True)

        # join based on objectid
        merged_df = input_copy_df.merge(enrich_result_df, on=orig_index_field)
        merged_df.drop([orig_index_field], axis=1, inplace=True)

        # rearrange columns to move SHAPE to the last one
        orig_cols = merged_df.columns.tolist()
        new_cols = [c for c in orig_cols if c != in_sedf.spatial.name] + [
            in_sedf.spatial.name
        ]
        final_df = merged_df[new_cols]

        # return new SeDF based on final_df; shape column name is the same as before
        final_df.spatial.set_geometry(in_sedf.spatial.name)
        return final_df

    def enrich(
        self,
        geographies: Union[pd.DataFrame, Geometry, Iterable, Path],
        enrich_variables: Union[pd.DataFrame, Iterable],
        proximity_type: Optional[str] = None,
        proximity_value: Optional[Union[float, int]] = None,
        proximity_metric: Optional[str] = None,
        return_geometry: bool = True,
        output_spatial_reference: Union[int, dict, SpatialReference] = None,
        estimate_credits: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """
        Enrich enables retrieving apportioned demographic factors for input geographies.

        =============================       ====================================================================
        **Parameter**                        **Description**
        -----------------------------       --------------------------------------------------------------------
        geographies                         Input geographies desired to get demographic variables for. Normally
                                            these will be geometries included as part of a spatially enabled Pandas
                                            Data Frame, but if using standard geographies, this can also be just an
                                            iterable of standard geography identifiers.
        -----------------------------       --------------------------------------------------------------------
        enrich_variables                    Variables desired to be retrieved.
        -----------------------------       --------------------------------------------------------------------
        proximity_type                      If the input geographies are points, retrieving enriched
                                            variables requires delineating a zone around each point to use for apportioning
                                            demographic factors to each input geography. Default is ``straight_line``.
        -----------------------------       --------------------------------------------------------------------
        proximity_value                     If the input geographies are points, this is the value used
                                            to create a zone around the points for apportioning demographic factors. For
                                            instance, if specifying five miles, this parameter value will be ``5``. Default
                                            is ``1``.
        -----------------------------       --------------------------------------------------------------------
        proximity_metric                    If the input geographies are points, this is the metric
                                            defining the proximity value. For instance, if specifying one kilometer, this
                                            value will be ``kilometers``. Default is ``kilometers``.
        -----------------------------       --------------------------------------------------------------------
        return_geometry                     Whether or not it is desired to have geometries returned as part
                                            of the returned dataframe. Default is ``True``.
        -----------------------------       --------------------------------------------------------------------
        output_spatial_reference            If the geometry is being returned, and a geometry other
                                            than WGS84 is desired, please provide it here. The default is
                                            ``{'wkid': 4326}`` (WGS84).
        -----------------------------       --------------------------------------------------------------------
        estimate_credits                    If the source for the Business Analyst instance is ArcGIS Online,
                                            this enables estimation of credit consumption before actually performing the
                                            enrich task.
        =============================       ====================================================================

        Returns:
            Pandas Data Frame
        """
        from arcgis.geoenrichment.enrichment import NamedArea

        # pull out country specific parameters from the kwargs
        country, kwargs = extract_from_kwargs("country", kwargs)
        standard_geography_level, kwargs = extract_from_kwargs(
            "standard_geography_level", kwargs
        )
        standard_geography_id_column, kwargs = extract_from_kwargs(
            "standard_geography_id_column", kwargs
        )

        # if the geographies input is just a single geometry, convert to list to work with
        if isinstance(geographies, Geometry):
            geographies = [geographies]

        # if a dict is passed directly in, check if it is Geometry or a FeatureSet
        if isinstance(geographies, dict):
            if is_dict_geometry(geographies):
                geographies = Geometry(geographies)
            elif is_dict_featureset(geographies):
                geographies = FeatureSet(geographies)
            # dict of named areas
            elif isinstance(list(geographies.values())[0], NamedArea):
                named_areas = list(geographies.values())
                geographies = [named_area.geometry for named_area in named_areas]

        if isinstance(geographies, Iterable) and not isinstance(
            geographies, pd.DataFrame
        ):
            # if a list of geometries is passed in dict form, convert to Geometry objects
            if isinstance(geographies[0], dict):
                if is_dict_geometry(geographies[0]):
                    geographies = [Geometry(g_dict) for g_dict in geographies]

        # if a FeatureSet, convert to spatially enabled data frame
        if isinstance(geographies, FeatureSet):
            geographies = geographies.sdf

        # flag if a dataframe
        geo_is_df = isinstance(geographies, pd.DataFrame)

        # flag if is list of dictionaries, which allowed through untouched
        geo_is_dict = False
        if isinstance(geographies, Iterable):
            first_geo = (
                geographies.iloc[0]
                if isinstance(geographies, pd.DataFrame)
                else geographies[0]
            )
            if isinstance(first_geo, dict):
                geo_is_dict = True

        # check if a spatially enabled dataframe, if standard geography identifiers are not provided
        if geo_is_df and standard_geography_id_column is None and not geo_is_dict:
            assert geographies.spatial.validate(), (
                "If providing a Pandas DataFrame for enrichment, you must either "
                "ensure it is a valid spatially enabled dataframe or provide a "
                "column containing standard geography identifiers."
            )

        # if a dataframe and the geography id column is provided, pluck out the standard geography identifiers
        elif geo_is_df and standard_geography_id_column and not geo_is_dict:
            geographies = geographies[standard_geography_id_column]

        if geo_is_df and output_spatial_reference is None:
            output_spatial_reference = geographies.spatial.sr
        elif geo_is_dict and output_spatial_reference is None:
            if "spatialReference" in first_geo:
                output_spatial_reference = first_geo["spatialReference"]
        # ensure if specifying a standard geography id column, the standard geography level is also provided
        if standard_geography_id_column is not None:
            assert standard_geography_level is not None, (
                "If providing a standard_geography_id_column, you also need "
                "to provide a standard_geography_level."
            )
            assert (
                country is not None
            ), "If using a standard geography, you must also provide a country."

        # make sure the standard geography level is either an integer index or recognized string
        if isinstance(standard_geography_level, int):
            msg_cntry_idx = (
                "If providing an integer index for the standard_geography_level, it must be within "
                "the available range."
            )
            assert (
                standard_geography_level in country.geography_levels.index
            ), msg_cntry_idx

        elif isinstance(standard_geography_level, str):
            # crawl through all the available columns looking for a match and retrieve the index
            lvl_mtch = False
            for c in country.geography_levels.columns:
                std_geo_lwr = standard_geography_level.lower()
                mtch_srs = (
                    country.geography_levels[c].str.lower().str.contains(std_geo_lwr)
                )
                if mtch_srs.any():
                    lvl_mtch = True
                    standard_geography_level = country.geography_levels[mtch_srs].index[
                        0
                    ]
                    break

            # make sure the geography level was found
            assert lvl_mtch, (
                "If providing a standard_geography_level, it must match one of the available "
                "geography levels."
            )

        # if the geographies is not a path and not a dataframe, convert the iterable to a list for consistency later
        if not isinstance(geographies, (pd.DataFrame, Path)) and not isinstance(
            geographies, str
        ):
            geographies = list(geographies)

        # get enrichment variables to validate against depending on the enrichment variable source
        ev_df = (
            country.enrich_variables if country is not None else self.enrich_variables
        )

        # if no variables submitted, provide defaults flexibly based on the parent
        if enrich_variables is None:
            # pluck out enrich variables into a shorter variable
            ev = self.enrich_variables

            # get the current year key variables
            enrich_variables = (
                ev[
                    (ev.name.str.lower().str.contains("cy"))
                    & (ev.data_collection.str.lower().str.contains("key"))
                ]
                .drop_duplicates("name")
                .reset_index(drop=True)
            )

            # ensure something is found, dropping current year if nothing found
            if len(enrich_variables.index) == 0:
                enrich_variables = (
                    ev[(ev.data_collection.str.lower().str.contains("key"))]
                    .drop_duplicates("name")
                    .reset_index(drop=True)
                )

        # if a list of enrichment variables was provided, ensure they are valid
        if not isinstance(enrich_variables, pd.DataFrame):
            # iteratively go through all the columns and try to find matching variables
            for c in ev_df:
                enrich_vars = [v.lower() for v in enrich_variables]
                e_df = ev_df[ev_df[c].str.lower().isin(enrich_vars)].reset_index(
                    drop=True
                )
                if len(e_df.index):
                    break

            # pitch a fit if no variables were found
            if len(e_df.index) == 0:
                raise ValueError(
                    "None of the provided enrich_variables appear to be available."
                )

            # provide a message if not all enrich_variables were found
            if len(enrich_variables) > len(e_df.index):
                warn(
                    f"Not all input enrich_variables were matched ({len(e_df.index):,}/{len(enrich_variables):,})"
                )

        # set the default proximity metric if none provided
        if proximity_metric is None:
            proximity_metric = "kilometers"

        e_df = self._enrich(
            geographies,
            enrich_variables,
            country,
            standard_geography_level,
            proximity_type,
            proximity_value,
            proximity_metric,
            return_geometry,
            output_spatial_reference,
            estimate_credits,
            **kwargs,
        )

        return e_df

    @local_vs_gis
    def _enrich(
        self,
        geographies: Union[pd.DataFrame, Geometry, Iterable, Path],
        enrich_variables: pd.DataFrame,
        country: Optional[Country] = None,
        standard_geography_level: Optional[Union[int, str]] = None,
        proximity_type: str = "straight_line",
        proximity_value: Union[float, int] = 1,
        proximity_metric: str = "Kilometers",
        return_geometry: bool = True,
        output_spatial_reference: Union[int, dict, SpatialReference] = 4326,
        estimate_credits: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Redirect enabling all the front end validation to be handled in the non-underscore function."""
        pass

    def _enrich_local(
        self,
        geographies: Union[pd.DataFrame, Geometry, Iterable, Path],
        enrich_variables: pd.DataFrame,
        country: Optional[Country] = None,
        standard_geography_level: Optional[Union[int, str]] = None,
        proximity_type: Optional[str] = None,
        proximity_value: Optional[Union[float, int]] = None,
        proximity_metric: Optional[str] = None,
        return_geometry: bool = True,
        output_spatial_reference: Union[int, dict, SpatialReference] = 4326,
        estimate_credits: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Local enrich implementation."""
        assert estimate_credits is False, (
            "Cannot estimate credits for local analysis since using a local source does "
            "not use credits."
        )

        # if unsupported local analysis parameters provided, error
        unsupported_params = [
            "comparison_levels",
            "add_derivative_variables",
            "intersecting_geographies",
        ]
        invalid_params = [k for k in kwargs.keys() if k in unsupported_params]
        if len(invalid_params):
            raise ValueError(
                f"{', '.join(invalid_params)} input parameters are not supported with local analysis."
            )

        # also, raw json input is also not supported
        if isinstance(geographies, (Iterable, pd.Series)) and not isinstance(
            geographies, pd.DataFrame
        ):
            if not isinstance(geographies[0], Geometry) and isinstance(
                geographies[0], dict
            ):
                raise ValueError(
                    "Raw JSON is not supported as input for geoenrichment when using a 'local' source."
                )

        # late import to avoid collisions
        import arcpy

        # ensure no conflicts with intermediate datasets
        arcpy.env.overwriteOutput = True

        # get the variables property formatted for the Enrich Layer tool
        if country is not None:
            evars = self._enrich_variable_preprocessing(
                enrich_variables, country=country
            )
        else:
            evars = self._enrich_variable_preprocessing(enrich_variables)

        # if the standard geographies were provided
        if standard_geography_level is not None:
            # if input is a dataframe, ensure correctly set up
            if isinstance(geographies, pd.DataFrame):
                std_geo_in = get_spatially_enabled_dataframe(
                    geographies
                ).spatial.to_featureclass("memory/tmp_in_geo")

            # if input is any other iterable or Series, convert to comma separated string
            elif isinstance(geographies, (Iterable, pd.Series)):
                std_geo_in = ", ".join(list(geographies))

            # convert the standard geography to the correct string for use in later geoprocessing
            standard_geography_level = country.geography_levels.iloc[
                standard_geography_level
            ]["level_id"]

            # retrieve the standard geographies based on the identifiers
            geographies = arcpy.ba.StandardGeographyTA(
                geography_level=standard_geography_level,
                out_feature_class="memory/tmp_std_geo",
                input_type="LIST",
                ids_list=std_geo_in,
            )[0]

            # convert to spatially enabled dataframe
            geographies = GeoAccessor.from_featureclass(geographies)

        # otherwise, make sure input in consistent format
        else:
            geographies = get_spatially_enabled_dataframe(geographies)

            # if z-enabled, de-enable so enrich can work...conversion does not work with z-enabled features
            if geographies[geographies.spatial.name].iloc[0].has_z:
                geographies[geographies.spatial.name] = geographies[
                    geographies.spatial.name
                ].apply(lambda geom: Geometry(geom.__geo_interface__))

        # if a proximity type is provided, validate
        if proximity_type is not None:
            proximity_type = validate_network_travel_mode(
                country, proximity_type, proximity_metric
            )

            # provide defaults if nothing provided for any of the proximity metrics
            proximity_type = (
                "Straight Line" if proximity_type is None else proximity_type
            )
            proximity_metric = (
                "kilometers" if proximity_metric is None else proximity_metric
            )
            proximity_value = 1 if proximity_value == 1 else proximity_value

            # ensure if the geometry is lines, the proximity_type is not a network travel mode
            if (
                geographies.spatial.geometry_type[0] == "polyline"
                and proximity_type is not None
            ):
                assert (
                    proximity_type == "Straight Line"
                ), 'Only "Straight Line" proximity_type can be used with line geometry.'

        # handle pre 2.9
        in_geo = (
            geographies
            if pro_at_least_version("2.9")
            else geographies.spatial.to_featureclass(
                f"memory/geo_in_{uuid.uuid4().hex}"
            )
        )

        # now, actually perform enrichment
        use_arrow = pro_at_least_version("3.1") and self._can_use_arrow(in_geo)

        if use_arrow:
            enrich_res = self._enrich_using_arrow(in_sedf=in_geo, variables=evars)
        else:
            enrich_res = arcpy.ba.EnrichLayer(
                in_features=in_geo,
                out_feature_class=f"memory/tmp_enrich_{uuid.uuid4().hex}",
                variables=evars,
                buffer_type=proximity_type,
                distance=proximity_value,
                unit=proximity_metric,
            )

        # handle differences in pre 2.9
        if not pro_at_least_version("2.9"):
            arcpy.management.Delete(in_geo)
            enrich_res = enrich_res[0]

        # cache output depending on what the input was
        enrich_res = (
            enrich_res if isinstance(geographies, pd.DataFrame) else enrich_res[0]
        )

        # convert the output to a spatially enabled dataframe if necessary (in 3.1 it's done in
        enrich_df = (
            enrich_res if use_arrow else GeoAccessor.from_featureclass(enrich_res)
        )

        if not use_arrow:
            # in some cases from_featureclass returns a data frame that doesn't pass SEDF validation (enrich_df.spatial.validate)
            enrich_df.spatial.set_geometry("SHAPE")

        # standardize columns to ensure results are as expected
        enrich_df.columns = [
            self._standardize_enrich_column_name(c, country) for c in enrich_df.columns
        ]

        column_candidates_to_remove = ["Shape_Area", "Shape_Length"]
        # default value set to True to keep backward compatibility
        sanitize_columns = kwargs.pop("sanitize_columns", True)
        if sanitize_columns:
            enrich_df.columns = [
                pep8ify(c) for c in enrich_df.columns if c != "SHAPE"
            ] + ["SHAPE"]
            column_candidates_to_remove = [
                pep8ify(c) for c in column_candidates_to_remove
            ]

        # start creating a list of columns to remove - beginning with the OBJECTID field
        drop_cols = [c for c in enrich_df.columns if c in column_candidates_to_remove]

        if not use_arrow:
            if sanitize_columns:
                drop_cols.append(pep8ify(arcpy.Describe(enrich_res).OIDFieldName))
            else:
                drop_cols.append(arcpy.Describe(enrich_res).OIDFieldName)

        if not use_arrow:
            # get rid of the temporary output to save memory
            arcpy.management.Delete(enrich_res)

        # if returning geometry
        if return_geometry:
            # ensure the spatial reference is correct

            if output_spatial_reference:
                enrich_df = change_spatial_reference(
                    enrich_df, output_spatial_reference
                )

            # removing unneeded columns - using inplace to preserve all spatial namespace properties
            enrich_df.drop(columns=drop_cols, inplace=True)

        # if not returning geometry, drop the geometry column
        else:
            drop_cols.append(enrich_df.spatial.name)

            # remove unneeded columns - not doing inplace to ensure no 'spatial' namespace remnants
            enrich_df = enrich_df.drop(columns=drop_cols)

        return enrich_df

    def _enrich_gis(
        self,
        geographies: Union[pd.DataFrame, Geometry, Iterable, Path],
        enrich_variables: pd.DataFrame,
        country: Optional[Country] = None,
        standard_geography_level: Optional[Union[int, str]] = None,
        proximity_type: Optional[str] = None,
        proximity_value: Optional[Union[float, int]] = None,
        proximity_metric: Optional[str] = None,
        return_geometry: bool = True,
        output_spatial_reference: Union[int, dict, SpatialReference] = None,
        estimate_credits: bool = False,
        **kwargs,
    ) -> pd.DataFrame:
        """Web GIS implementation for _enrich"""
        # before going any further, make sure can enrich using current user (if any)
        if self.source.users.me is not None:
            has_ge = (
                len(
                    [p for p in self.source.users.me.privileges if "geoenrichment" in p]
                )
                > 0
            )
        elif (
            "appInfo" in self.source.properties
            and self.source.properties.appInfo.privileges
            and "premium:user:geoenrichment"
            in self.source.properties.appInfo.privileges
        ):
            has_ge = True
        else:
            has_ge = False
        assert has_ge, (
            "Cannot enrich due to insufficient permissions. Please ensure the GIS object instance is "
            "created with credentials, a user, with permissions to perform geoenrichment."
        )

        # TODO: implement estimate credits
        if estimate_credits:
            raise NotImplementedError("estimate_credits not yet implemented")

        # pitch a fit with a path
        if isinstance(geographies, Path):
            raise ValueError(
                "Providing a direct path to a table or feature class for enrichment is only supported "
                "when using a local ArcGIS Pro source."
            )

        # best practices batch size
        batch_size = 50

        # construct geoenrichment enrich method url
        ge_url = f"{self._base_url}/Geoenrichment/Enrich"

        # get the enrichment variables as a string ready to submit as a payload parameter
        evars = self._enrich_variable_preprocessing(enrich_variables, country=country)

        if output_spatial_reference is None:
            output_spatial_reference = 4326
        # properly format the output spatial reference
        if isinstance(output_spatial_reference, (int, str)):
            output_spatial_reference = SpatialReference(output_spatial_reference)

        # start building out the package for enrich REST call
        params = {
            "f": "json",
            "analysisVariables": evars,
            "appID": "esripythonapi",
        }

        # if any legacy parameters provided, add them to the payload
        legacy_params = [
            "comparison_levels",
            "add_derivative_variables",
            "intersecting_geographies",
        ]
        for param in [k for k in kwargs.keys() if k in legacy_params]:
            # convert the parameter key from snake case to camel case
            prts = param.split("_")
            param_cml = prts[0] + "".join(prt.title() for prt in prts[1:])

            # tack onto the payload
            params[param_cml] = kwargs[param]

        # if working with a specific country, add this to the payload
        if country is not None:
            hierarchy = kwargs.pop("hierarchy", country.properties.hierarchy[0])
            params["useData"] = json.dumps(
                {"sourceCountry": country.properties.iso3, "hierarchy": hierarchy}
            )

        # get the maximum batch size to ensure is not less than best practices set above
        svc_lmt_url = f'{self.source.properties.helperServices("geoenrichment").url}/Geoenrichment/ServiceLimits'
        svc_lmt_res = self.source._con.get(svc_lmt_url)

        max_batch_size = [
            v["value"]
            for v in svc_lmt_res["serviceLimits"]["value"]
            if v["paramName"] == "maximumStudyAreasNumber"
        ][0]
        batch_size = batch_size if max_batch_size > batch_size else max_batch_size

        # if a string, or list of strings, and no standard geography level is provided, format as address in JSON
        if isinstance(geographies, str) and standard_geography_level is None:
            geographies = [geographies]
        if isinstance(geographies, Iterable) and not isinstance(
            geographies, pd.DataFrame
        ):
            if isinstance(geographies[0], str) and standard_geography_level is None:
                geographies = [
                    {"address": {"text": addr_str}} for addr_str in geographies
                ]

        # convert boolean to string for payload in correct circumstances
        if return_geometry:
            retrieve_geometry = True
            params["returnGeometry"] = "true"
        else:
            retrieve_geometry = False
            params["returnGeometry"] = "false"

        # list to store request parameter payloads
        req_param_lst = []

        # detect and flag if a list of Geometry. String and dictionary objects passed directly in as iterable
        is_dict = False
        is_geom = False
        is_str = False
        if isinstance(geographies, list):
            if isinstance(geographies[0], Geometry):
                is_geom = True
            elif isinstance(geographies[0], dict):
                is_dict = True
            if isinstance(geographies[0], str):
                is_str = True

        # if working with standard geography
        if standard_geography_level is not None:
            assert (
                country is not None
            ), "Standard geography levels can only be used with a Country."

            # pull the geography level out of the geography levels dataframe
            geo_lvl = country.geography_levels.iloc[standard_geography_level][
                "level_id"
            ]

            # use the count of features and the max bach size to create a list of param payloads
            for idx in range(0, len(geographies), batch_size):
                # peel off just the id's for this batch
                batch_id_lst = geographies[idx : idx + batch_size]

                # create the param payload
                params["studyAreas"] = json.dumps(
                    [
                        {
                            "sourceCountry": country.iso3,
                            "layer": geo_lvl,
                            "ids": batch_id_lst,
                        }
                    ]
                )

                # add the payload to the list
                req_param_lst.append(deepcopy(params))

        # if a list of dictionaries, which are not geometries, or a list of strings is being passed in, just send
        elif (is_dict and not is_geom) or is_str:
            if proximity_value is not None:
                prx_src = self if country is None else country
                geographies = add_proximity_to_enrich_feature_list(
                    prx_src,
                    geographies,
                    proximity_type,
                    proximity_metric,
                    proximity_value,
                )
            for idx in range(0, len(geographies), batch_size):
                geo_btch = geographies[idx : idx + batch_size]
                params["studyAreas"] = json.dumps(geo_btch) if is_dict else geo_btch
                req_param_lst.append(deepcopy(params))

        # otherwise, working with geometries, so do this thing
        else:
            # no matter what the input, get a spatially enabled dataframe
            geographies = get_spatially_enabled_dataframe(geographies)

            # tack on the spatial reference to make sure it comes along for the ride
            params["insr"] = json.dumps(geographies.spatial.sr)

            # tack on the output spatial reference as well
            params["outsr"] = json.dumps(output_spatial_reference)

            # check to make sure a valid geometry is present
            geom_typ_lst = [
                gt for gt in geographies.spatial.geometry_type if isinstance(gt, str)
            ]
            assert len(
                geom_typ_lst
            ), "The Dataframe does not appear to have a valid geometry type."

            # make sure there is a unique identifier for combining data later if more than just geometries provided
            if isinstance(geographies, pd.DataFrame):
                geographies["enrich_idx"] = geographies.index

            # use the count of features and the max bach size to create a list of param payloads
            total_cnt = (
                len(geographies.index)
                if isinstance(geographies, pd.DataFrame)
                else len(geographies)
            )
            for idx in range(0, total_cnt, batch_size):
                # get a slice of the input data to enrich, converting to feature set if dataframe
                if isinstance(geographies, pd.DataFrame):
                    in_batch_df = geographies.iloc[idx : idx + batch_size]

                    # format the features for sending - keep it light, just the geometry
                    batch_df = in_batch_df[[in_batch_df.spatial.name, "enrich_idx"]]
                    batch_df.spatial.set_geometry(geographies.spatial.name)
                    batch_features = batch_df.spatial.to_featureset().features
                    feature_lst = [f.as_dict for f in batch_features]

                    # if proximity metrics provided, add the proximity metrics to the list of features
                    if proximity_value is not None:
                        prx_src = self if country is None else country
                        feature_lst = add_proximity_to_enrich_feature_list(
                            prx_src,
                            feature_lst,
                            proximity_type,
                            proximity_metric,
                            proximity_value,
                        )

                # ...and if already list of dict objects, split into parts
                else:
                    feature_lst = geographies[idx : idx + batch_size]

                # convert everything to string for payload
                params["studyAreas"] = json.dumps(feature_lst)

                # add the payload onto the list
                req_param_lst.append(deepcopy(params))

        # bach request asynchronously
        enrich_res_df = run_async(
            _get_enrich_rest, ge_url, req_param_lst, retrieve_geometry, self.source
        )

        # clean up the response dataframe schema
        drop_cols = [c for c in enrich_res_df.columns if "objectid" in c.lower()] + [
            "ID"
        ]
        enrich_res_df.drop(columns=drop_cols, inplace=True, errors="ignore")

        # if more than just a list of standard geography id's was the input, combine the input with the results
        if isinstance(geographies, pd.DataFrame):
            src_drop_cols = [geographies.spatial.name] + [
                c for c in geographies.columns if c.lower().startswith("shape_")
            ]
            enrich_df = enrich_res_df.join(
                geographies.drop(columns=src_drop_cols).set_index("enrich_idx"),
                on="enrich_idx",
            ).drop(columns="enrich_idx")
            cols_in_order = [
                c for c in geographies.columns if c not in src_drop_cols
            ] + list(enrich_res_df.columns)
            enrich_df = enrich_df[[c for c in cols_in_order if c != "enrich_idx"]]
        else:
            enrich_df = enrich_res_df

        # free up memory
        del enrich_res_df

        # if geometry, which may not be, but if there is, clean up and make sure everything is as expected
        if "SHAPE" in enrich_df.columns:
            # ensure it is valid to begin with - doubtful after the join
            enrich_df.spatial.set_geometry("SHAPE")

            # shuffle columns so geometry is at the end
            enrich_df = enrich_df[
                [c for c in enrich_df.columns if c != "SHAPE"] + ["SHAPE"]
            ]

            # set the geometry
            enrich_df.spatial.set_geometry("SHAPE")

        # proactively change the column names so no surprises if exporting to a feature class later
        enrich_df.columns = [
            self._standardize_enrich_column_name(c, country) for c in enrich_df.columns
        ]

        # default value set to True to keep backward compatibility
        sanitize_columns = kwargs.pop("sanitize_columns", True)
        if sanitize_columns:
            enrich_df.columns = [
                pep8ify(c) if c != "SHAPE" else c for c in enrich_df.columns
            ]

        # stash useful pieces for potential later access in metadata
        enrich_df.attrs["arcgis_ba"] = self
        enrich_df.attrs["arcgis_aoi"] = country

        return enrich_df

    @lru_cache(maxsize=255)
    def _standardize_enrich_column_name(
        self, column_name: str, country: Optional[Country] = None
    ):
        """Helper function to standardize the output column names so is the same no matter the source."""
        std_src = self if country is None else country
        col_nm = std_src.get_enrich_variables_from_iterable(
            column_name, suppress_warn=True
        )
        col_nm = col_nm.iloc[0]["name"] if len(col_nm.index) > 0 else column_name
        return col_nm


async def _get_enrich_rest(
    ge_url: str, payload_lst: Iterable[dict], retrieve_geometry: bool, source: GIS
) -> Awaitable[pd.DataFrame]:
    """Function enabling batching of enrich rest call asynchronously."""
    # variable for storing results
    enrich_res_itr = []

    # iterate the batched payloads
    for payload in payload_lst:
        # create an event loop for wrapping requests
        loop = asyncio.get_event_loop()

        # get a listener, a future object, and send request to the server
        future = loop.run_in_executor(None, source._con.post, ge_url, payload)

        # hold short for response (but since using async, other requests get queued up)
        res = await future

        # ensure a valid result is received
        if "error" in res["messages"]:
            err = res["messages"]["error"]
            raise Exception(
                "Error in enriching data using Business Analyst Enrich REST endpoint - Error "
                f'Code {err["code"]}: {err["message"]}'
            )

        # pull out the response feature set
        fs = res["results"][0]["value"]["FeatureSet"]
        assert (
            len(fs) > 0
        ), "No results were returned. Please ensure you are using the correct country."

        # if getting geometry back, unpack into spatially enabled dataframe
        if retrieve_geometry:
            r_df = FeatureSet.from_dict(fs[0]).sdf

        # unpack the enriched results - reaching into the FeatureSet for just the attributes - much faster
        else:
            r_df = pd.DataFrame([f["attributes"] for f in fs[0]["features"]])

        # add the dataframe to the list
        enrich_res_itr.append(r_df)

    # combine all the received enriched data and take out the trash
    res_df = pd.concat(enrich_res_itr).reset_index(drop=True)
    del enrich_res_itr

    return res_df



from __future__ import annotations
from arcgis.gis import GIS


class MarketPlaceManager:
    """Provides the ability for the manager to list and unlist marketplace items"""

    _gis: GIS = None
    _url: str = None

    def __init__(self, gis: GIS):
        self._gis = gis
        self._url = f"{gis._portal.resturl}content"

    # ----------------------------------------------------------------------
    def list(self, itemid: str) -> dict:
        """
        This operation lists the item in the marketplace.

        This operation is only available to organizations that have permissions
        to list items in the marketplace. The permissions are returned with the
        Portal Self response.

        The listing properties must be specified for the item before
        calling this operation. This operation will fail if listing
        properties have not already been specified.

        Listing an item will set its listed property to true.

        This operation is available to the user and to the administrator
        of the organization to which the user belongs.

        :return: A dictionary indicating success or failure and the item id of the listed item.
        """
        params = {"f": "json"}
        url = f"{self._url}/users/{self._gis.users.me.username}/items/{itemid}/list"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def unlist(self, itemid: str) -> dict:
        """
        This operation unlists a previously listed item from the marketplace.

        Unlisting an item will reset its listed property to false.

        This operation is available to the user and the administrator of
        the organization to which the user belongs.

        :return: A dictionary indicating success or failure and the item id of the listed item.
        """
        params = {"f": "json"}
        url = f"{self._url}/users/{self._gis.users.me.username}/items/{itemid}/unlist"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def listings(
        self,
        query: str = "*",
        my_listings: bool = False,
        start: int = 1,
        num: int = 10,
        sort_field: str | None = None,
        sort_order: str = "asc",
    ) -> dict:
        """
        This operation searches for marketplace listings. The searches are performed
        against a high performance index that indexes the most popular fields of a listing.

        By default, this search spans all public listings in the marketplace. However,
        if you're logged in as a vendor org admin and you specify the my_listings=true parameter,
        it then searches all public and private listings in your organization.


        ======================      ====================================================================
        **Parameter**                **Description**
        ----------------------      --------------------------------------------------------------------
        query                       Optional string. The query string to use to search.
        ----------------------      --------------------------------------------------------------------
        my_listings                 Optional boolean.  If True and you're logged in as
                                    a vendor org admin, it searches all public and private
                                    listings in your organization. Note that if my_listings=True,
                                    the query parameter is optional. Default is False.
        ----------------------      --------------------------------------------------------------------
        start                       Optional integer. The number of the first entry in the
                                    result set response. The index number is 1-based.

                                    The default value of start is 1. (i.e., the first search result).

                                    The start parameter, along with the num parameter,
                                    can be used to paginate the search results.
        ----------------------      --------------------------------------------------------------------
        num                         Optional integer. The maximum number of results to be
                                    included in the result set response.

                                    The default value is 10, and the maximum allowed value is 100.

                                    The start parameter, along with the num parameter,
                                    can be used to paginate the search results.

                                    Note that the actual number of returned results may be
                                    less than num. This happens when the number of results
                                    remaining after start is less than num.
        ----------------------      --------------------------------------------------------------------
        sort_field                  Optional string. The field to sort by. You can also sort
                                    by multiple fields (comma separated) for listings, sort
                                    field names are case-insensitive.

                                    Supported sort field names are:
                                    "title", "created", "listingpublisheddate", "type", "owner",
                                    "avgrating", "numratings", "numcomments", and "numviews"
        ----------------------      --------------------------------------------------------------------
        sort_order                  Optional string. Describes whether the order returns
                                    in ascending(asc) or descending(desc) order. Default is asc.
        ======================      ====================================================================


        :return:
            A dictionary with response syntax of:

                | {
                | "query": "<query string>",
                | "total": <total number of results>,
                | "start": <results in first set>,
                | "num": <number of results per page>,
                | "nextStart": <result number of next page>,
                | "listings": [{<listing1>}, {<listing2>}]
                | }

        """
        url = f"{self._url}/listings"
        params = {
            "f": "json",
            "q": query,
            "mylistings": my_listings,
            "start": start,
            "num": num,
            "sortField": sort_field,
            "sortOrder": sort_order,
        }
        return self._gis._portal.con.get(url, params)

    # ----------------------------------------------------------------------
    def listing(self, itemid: str) -> dict:
        """
        A listing in the marketplace. The listing and its corresponding item share the same ID.

        =====================       ========================================
        **Parameter**                **Description**
        ---------------------       ----------------------------------------
        itemid                      Required String. The item id.
        =====================       ========================================

        :return:
            A dictionary of the listed item with properties.
        """
        params = {"f": "json"}
        url = f"{self._url}/listings/{itemid}"
        return self._gis._portal.con.get(url, params)

    # ----------------------------------------------------------------------
    def delete_provision(self, itemid: str) -> dict:
        """
        This operation deletes all provisions to this item for
        the specified purchaser.

        This operation cannot be invoked if the item has not been provisioned
        to the specified purchaser.

        .. note::
            Only vendor org admins can invoke this operation.

        =====================       ========================================
        **Parameter**                **Description**
        ---------------------       ----------------------------------------
        itemid                      Required String. The item id.
        =====================       ========================================

        :return:
            A dictionary with syntax:

                | {
                | "success": <true | false>,
                | "itemId": "<itemId>",
                | "purchaserOrgId": "<purchaserOrgId>"
                | }

        """
        params = {"f": "json"}
        url = f"{self._url}/listings/{itemid}/deleteProvision"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def express_interest(self, itemid: str) -> dict:
        """
        A purchaser can express interest in a marketplace listing by
        invoking this operation.

        This operation cannot be invoked if the item has already been
        purchased or if the purchaser has previously expressed interest.

        Only administrators and members with request purchase information
        privilege of purchasing orgs can invoke this operation.

        Note that interests cannot be expressed for free listings, because
        they can be directly purchased by purchasing org admins or members
        with request purchase information privilege.

        =====================       ========================================
        **Parameter**                **Description**
        ---------------------       ----------------------------------------
        itemid                      Required String. The item id.
        =====================       ========================================

        :return:
            A dictionary of the listed item with properties.
        """
        params = {"f": "json"}
        url = f"{self._url}/listings/{itemid}/interest"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def provision_org_entitlements(
        self,
        itemid: str,
        purchaser_org_id: str,
        purchaser_subscription_id: str | None = None,
        org_entitlements: dict | None = None,
    ) -> dict:
        """
        For a license-by-user listing, selling organization administrator
        or members can use this operation to provision entitlements to a
        purchasing organization. It can only be made if the item has already
        been purchased, or is being tried by the purchasing org.

        This operation is HTTPS only for Esri apps that require a signature,
        otherwise it can be either for provider apps.

        It can only be invoked by org admins or members with request
        purchase information privilege.

        ==========================      ==================================================================================================================================================================
        **Parameter**                    **Description**
        --------------------------      ------------------------------------------------------------------------------------------------------------------------------------------------------------------
        itemid                          Required String. The item id.
        --------------------------      ------------------------------------------------------------------------------------------------------------------------------------------------------------------
        purchaser_org_id                Required String. The org ID of the purchasing organization
        --------------------------      ------------------------------------------------------------------------------------------------------------------------------------------------------------------
        purchaser_subscription_id       Optional String. The subscription(SMS) ID of the purchasing organization.
        --------------------------      ------------------------------------------------------------------------------------------------------------------------------------------------------------------
        org_entitlements                Required Dictionary. A JSON object representing the set of entitlements available to the purchasing org.

                                        Example:

                                            | {
                                            | "maxUsers": 10,
                                            | "entitlements": {
                                            | "standard": {"num": 8}, ('standard' is an entitlement string that uniquely identifies entitlement, listingID is used typically for provider apps)
                                            | "advanced": {"num": 2},
                                            | "spatialAnalyst": {"num": 2}
                                            | }
                                            | }

        ==========================      ==================================================================================================================================================================

        :return: A dictionary of the provision item.
        """
        params = {
            "f": "json",
            "purchaserOrgId": purchaser_org_id,
            "purchaserSubscriptionId": purchaser_subscription_id,
            "orgEntitlements": org_entitlements,
        }
        url = f"{self._url}/listings/{itemid}"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def provision_user_entitlements(self, itemid: str, user_entitlements: dict) -> bool:
        """
        For a license-by-user listing, purchasing organization administrator
        can use this operation to provision entitlements to org
        members. It can only be made if the item has already been purchased,
        or is being tried by the purchasing org. A maximum of 25 users can
        be provisioned in one request.

        =====================       ====================================================================================================================================================================================
        **Parameter**                **Description**
        ---------------------       ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
        itemid                      Required String. The item id.
        ---------------------       ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
        user_entitlements           Required Dictionary. A JSON object representing the set of entitlements
                                    assigned to the specified set of users.

                                    Example:

                                        | {
                                        | "users": ["username1", "username2"],
                                        | "entitlements": ["standard", "networkAnalyst"] ("standard" is an entitlement string that uniquely identifies entitlement, listing itemId is used typically for provider apps)
                                        | }

                                    Only members of the purchasing org can be specified in the request.

                                    Specified entitlements are assigned to all specified users. If different sets of
                                    entitlements are to be assigned to different users, multiple requests
                                    with this operation are required.

                                    When there is no entitlements specified, it will revoke access to
                                    the item completely for the specified users.

                                    The total number of currently provisioned users plus users specified in requests
                                    should be no larger than the maximum number of users allowed for the purchasing org.
        =====================       ====================================================================================================================================================================================

        :return:
            A boolean indicating success (True) or failure (False).
        """
        params = {"f": "json", "userEntitlements": user_entitlements}
        url = f"{self._url}/listings/{itemid}"
        res = self._gis._portal.con.post(url, params)
        return res["success"]

    # ----------------------------------------------------------------------
    def purchases(self, status: str = "active") -> dict:
        """
        The purchases resource returns a list of purchases, trials, and
        interests expressed by this organization for items in the marketplace.

        =====================       ====================================================================================
        **Parameter**                **Description**
        ---------------------       ------------------------------------------------------------------------------------
        status                      Optional String. Status of the listings to be returned. The default value is active.

                                    Accepted values are:
                                        * active: Only listings that are currently active will be returned
                                        * expired: Only listings that have already expired will be returned
                                        * all: Both active and expired listings will be returned
        =====================       ====================================================================================

        :return: A dictionary depicting the purchases, trials, and interests.
        """
        params = {"f": "json", "status": status}
        url = f"{self._gis._portal.resturl}portals/self/purchases"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def purchase(
        self,
        itemid: str,
        purchase_org_id: str | None = None,
        provisioned_itemid: str | None = None,
        end_date: str | None = None,
    ) -> dict:
        """
        =====================       ====================================================================================
        **Parameter**                **Description**
        ---------------------       ------------------------------------------------------------------------------------
        itemid                      Required String. The item id.
        ---------------------       ------------------------------------------------------------------------------------
        purchase_org_id             Required String. The org ID of the purchaser organization. This parameter is
                                    required only when the call is made by the vendor. It is ignored otherwise.
        ---------------------       ------------------------------------------------------------------------------------
        provisioned_itemid          Required String. The ID of the item to be provisioned if different from the one
                                    listed.

                                    Note that the listed item and the provisioned item must be related by the
                                    `Listed2Provisioned` relationship otherwise it will result in an error.

                                    This parameter is allowed only when the call is made by the vendor.
                                    It is ignored otherwise.
        ---------------------       ------------------------------------------------------------------------------------
        end_date                    Required String. The end/expiry date of this purchase if any. If this parameter is
                                    not specified, it implies an unexpiring purchase. The end date specified
                                    should be in milliseconds from epoch.
        =====================       ====================================================================================

        :return: A dictionary of the provision item.
        """

        params = {
            "f": "json",
            "purchaseOrgId": purchase_org_id,
            "provisionedItemId": provisioned_itemid,
            "endDate": end_date,
        }
        url = f"{self._url}/listings/{itemid}/purchase"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def trial(self, itemid) -> dict:
        """
        A purchaser can start a trial for a marketplace listing by invoking this operation.

        This operation is only supported for listings that support trials
        (whose trialSupported property is true). Once started, the trial will be valid
        for the duration of the trial specified on the listing (the trialDuration property).

        This operation cannot be invoked if the item has already been purchased or
        if the purchaser has started a trial previously.

        Only admins or members with request purchase information privilege of
        purchasing orgs can invoke this operation.

        =====================       =================================
        **Parameter**                **Description**
        ---------------------       ---------------------------------
        itemid                      Required String. The item id.
        =====================       =================================

        :return: A dictionary of the provision item.
        """
        params = {"f": "json"}
        url = f"{self._url}/listings/{itemid}/trial"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def comments(self, itemid: str):
        """Lists all comments for the item"""
        params = {"f": "json"}
        url = f"{self._url}/listings/{itemid}/comments"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def user_entitlements(self, itemid: str) -> dict:
        """
        This operation allows purchasing organization administrators or a
        user with Manage Licenses privilege to retrieve all user entitlements
        assigned to users in their organization.

        =====================       ==============================
        **Parameter**                **Description**
        ---------------------       ------------------------------
        itemid                      Required String. The item id.
        =====================       ==============================

        :return: A JSON document representing the set of entitlements assigned to the specified set of users.
        """
        params = {"f": "json"}
        url = f"{self._url}/listings/{itemid}/userEntitlements"
        try:
            return self._gis._portal.con.post(url, params)
        except:
            return None

    # ----------------------------------------------------------------------
    def user_entitlement(self, itemid: str, username: str) -> dict:
        """
        This resource allows user, purchasing organization administrator, and
        members with the manage licenses privilege to retrieve entitlements assigned to the user.

        =====================       ==============================
        **Parameter**                **Description**
        ---------------------       ------------------------------
        itemid                      Required String.
        ---------------------       ------------------------------
        username                    Required String.
        =====================       ==============================
        """
        params = {"f": "json"}
        url = f"{self._url}/listings/{itemid}/userEntitlements/{username}"
        return self._gis._portal.con.post(url, params)

    # ----------------------------------------------------------------------
    def customer_list(
        self,
        itemid: str | None = None,
        orgname: str | None = None,
        status: str = "all",
        type: str = "PURCHASE",
        modified: str | None = None,
        sort_fields: str | None = None,
        sort_order: str = "asc",
        include_listing: bool = True,
        num: int = 10,
        start: int = 1,
    ) -> dict:
        """
        The customers_list resource returns a list of purchases, trials, and
        interests expressed by customers for items listed by this organization
        in the marketplace. This operation allows filtering and sorting of provisions.

        =====================       ====================================================================================
        **Parameter**                **Description**
        ---------------------       ------------------------------------------------------------------------------------
        itemid                      Optional String. The item id of the provision to be returned.
        ---------------------       ------------------------------------------------------------------------------------
        orgname                     Optional String. Purchaser organization name of the provisions to be returned.
        ---------------------       ------------------------------------------------------------------------------------
        status                      Optional String. Status of the listings to be returned. The default value is active.

                                    Accepted values are:
                                        * active: Only listings that are currently active will be returned
                                        * expired: Only listings that have already expired will be returned
                                        * all: Both active and expired listings will be returned
        ---------------------       ------------------------------------------------------------------------------------
        type                        Optional String. Access type of the provisions to be returned:
                                        * REQUEST: Only provisions that have been requested will be returned.
                                        * TRIAL: Only trial provisions will be returned.
                                        * PURCHASE: Only subscription provisions will be returned.
                                        * REQUESTANDTRIAL: Both provisions that have been requested and trial provisions will be returned.
                                        * REQUESTANDPURCHASE: Both provisions that have been requested and subscriptions will be returned.
                                        * TRIALANDPURCHASE: Both trial provisions and subscriptions will be returned.

                                    Values: "REQUEST" | "TRIAL" | "PURCHASE" | "REQUESTANDTRIAL" | "REQUESTANDPURCHASE" | "TRIALANDPURCHASE"
        ---------------------       ------------------------------------------------------------------------------------
        modified                    Optional String. The last modified date of the provisions to be returned. The date
                                    specified should be in milliseconds from epoch.
        ---------------------       ------------------------------------------------------------------------------------
        sort_field                  Optional String. The fields to sort provisions by. The allowed sort field names are
                                    orgname, created, endDate, and modified.
        ---------------------       ------------------------------------------------------------------------------------
        sort_order                  Optional String. Describe whether the order returns in ascending (asc) or
                                    descending (desc) order. The default is asc.
        ---------------------       ------------------------------------------------------------------------------------
        include_listing             Optional Boolean. If True, listing objects are included in the provision response.
                                    The default is True.
        ---------------------       ------------------------------------------------------------------------------------
        num                         Optional Integer. The maximum number of provisions to be included in the result
                                    set response. The default value is 10, and the maximum allowed value is 100.
                                    The start parameter, along with the num parameter, can be used to paginate the
                                    query results. Note that the actual number of returned results may be less than num.
                                    This happens when the number of results remaining after start is less than num.
        ---------------------       ------------------------------------------------------------------------------------
        start                       Optional Integer. The number of the first entry in the result set response. The
                                    index number is 1-based. The default value of start is 1. (i.e. the first search result).
                                    The start parameter, along with the num parameter, can be used to paginate the
                                    query results.
        =====================       ====================================================================================

        :return: A dictionary.
        """
        params = {
            "f": "json",
            "status": status,
            "type": type,
            "itemId": itemid,
            "orgname": orgname,
            "modified": modified,
            "sortFields": sort_fields,
            "sortOrder": sort_order,
            "includeListing": include_listing,
            "num": num,
            "start": start,
        }
        url = f"{self._gis._portal.resturl}portals/self/customers"
        return self._gis._portal.con.get(url, params)

    # ----------------------------------------------------------------------
